"""FamBPM deviceSet."""

import sys
import time as _time
# from threading import Event as _Flag

import numpy as _np
from copy import deepcopy as _dcopy

from .device import DeviceSet as _DeviceSet
from ..search import BPMSearch as _BPMSearch
from ..namesys import SiriusPVName as _PVName
from .bpm import BPM


class FamBPMs(_DeviceSet):
    """Family of BPMs.

    Parameters
    ----------
        devname (str, optional)
            Device name. If not provided, defaults to DEVICES.SI.
            Determine the list of BPM names.
        bpmnames ((list, tuple), optional)
            BPM names list. If provided, it takes priority over 'devname'
            parameter. Defaults to None.
        ispost_mortem (bool, optional)
            Whether to control PM acquisition core. Defaults to False.
    """

    TIMEOUT = 10
    RFFEATT_MAX = 30
    PROPERTIES_ACQ = BPM.PROPERTIES_ACQ
    PROPERTIES_DEFAULT = BPM.PROPERTIES_DEFAULT
    ALL_MTURN_SIGNALS2ACQ = ('A', 'B', 'C', 'D', 'X', 'Y', 'Q', 'S')

    class DEVICES:
        """."""

        SI = 'SI-Fam:DI-BPM'
        BO = 'BO-Fam:DI-BPM'
        ALL = (BO, SI)

    def __init__(
            self, devname=None, bpmnames=None, ispost_mortem=False,
            props2init='all', mturn_signals2acq=('X', 'Y')):
        """."""
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')

        devname = _PVName(devname)
        bpm_names = bpmnames or _BPMSearch.get_names(
            filters={'sec': devname.sec, 'dev': devname.dev})
        self._ispost_mortem = ispost_mortem

        self._mturn_signals2acq = list(mturn_signals2acq)
        self.bpms = [BPM(
            dev, auto_monitor_mon=False, ispost_mortem=ispost_mortem,
            props2init=props2init) for dev in bpm_names]

        super().__init__(self.bpms[:], devname=devname)
        self._bpm_names = bpm_names
        self._csbpm = self.bpms[0].csdata
        self._initial_timestamps = None

        self._mturn_flags = dict()
        # NOTE: ACQCount-Mon need to be fixed on BPM's IOC
        # for bpm in devs:
        #     pvo = bpm.pv_object('ACQCount-Mon')
        #     pvo.auto_monitor = True
        #     self._mturn_flags[pvo.pvname] = _Flag()
        #     pvo.add_callback(self._mturn_set_flag)

    @property
    def bpm_names(self):
        """Return BPM names."""
        return self._bpm_names

    @property
    def csbpm(self):
        """Return control system BPM constants class."""
        return self._csbpm

    @property
    def mturn_signals2acq(self):
        """Return which signals will be acquired by get_mturn_signals."""
        return _dcopy(self._mturn_signals2acq)

    @mturn_signals2acq.setter
    def mturn_signals2acq(self, sigs):
        sigs = [s.upper() for s in sigs]
        diff = set(sigs) - set(self.ALL_MTURN_SIGNALS2ACQ)
        if diff:
            raise ValueError('The following signals do not exist: '+str(diff))
        self._mturn_signals2acq = sigs

    def set_attenuation(self, value=RFFEATT_MAX, timeout=TIMEOUT):
        """."""
        for bpm in self:
            bpm.rffe_att = value

        mstr = ''
        okall = True
        t0 = _time.time()
        for bpm in self:
            tout = timeout - (_time.time() - t0)
            if not bpm._wait('RFFEAtt-RB', value, timeout=tout):
                okall = False
                mstr += (
                    f'\n{bpm.devname:<20s}: ' +
                    f'rb {bpm.rffe_att:.0f} != sp {value:.0f}')

        print('RFFE attenuation set confirmed in all BPMs', end='')
        print(', except:' + mstr if mstr else '.')
        return okall

    def get_slow_orbit(self):
        """Get slow orbit vectors.

        Returns:
            orbx (numpy.ndarray, 160): Horizontal Orbit.
            orby (numpy.ndarray, 160): Vertical Orbit.

        """
        orbx, orby = [], []
        for bpm in self.bpms:
            orbx.append(bpm.posx)
            orby.append(bpm.posy)
        orbx = _np.array(orbx)
        orby = _np.array(orby)
        return orbx, orby

    def get_mturn_signals(self):
        """Get Multiturn signals matrices.

        Returns:
            list: Each component of the list is an numpy.ndarray with shape
                (N, 160), containing the values for the signals acquired.

        """
        sigs = [[] for _ in self._mturn_signals2acq]

        mini = int(sys.maxsize)  # a very large integer
        for bpm in self.bpms:
            for i, sn in enumerate(self._mturn_signals2acq):
                sn = 'sum' if sn == 'S' else sn.lower()
                name = 'mt_' + ('ampl' if sn in 'abcd' else 'pos') + sn
                sigs[i].append(getattr(bpm, name))
            mini = min(mini, _np.min([s[-1].size for s in sigs]))

        for i, sig in enumerate(sigs):
            for j, s in enumerate(sig):
                sig[j] = s[:mini]
            sigs[i] = _np.array(sig).T
        return sigs

    def get_mturn_timestamps(self):
        """Get Multiturn data timestamps.

        Returns:
            tsmps (numpy.ndarray, (160, N)): The i-th row has the timestamp of
                the i-th bpm for the N aquired signals.

        """
        tsmps = _np.zeros(
            (len(self.bpms), len(self._mturn_signals2acq)), dtype=float)
        for i, bpm in enumerate(self.bpms):
            for j, s in enumerate(self._mturn_signals2acq):
                s = 'SUM' if s == 'S' else s
                pvo = bpm.pv_object(f'GEN_{s}ArrayData')
                tv = pvo.get_timevars(timeout=self.TIMEOUT)
                tsmps[i, j] = pvo.timestamp if tv is None else tv['timestamp']
        return tsmps

    def get_sampling_frequency(self, rf_freq: float, acq_rate='') -> float:
        """Return the sampling frequency of the acquisition.

        Args:
            rf_freq (float): RF frequency.
            acq_rate (str, optional): acquisition rate. Defaults to ''.
            If empty string, it gets the configured acq. rate on BPMs

        Returns:
            float: acquisition frequency.

        """
        fs_bpms = {
            dev.get_sampling_frequency(rf_freq, acq_rate)
            for dev in self.bpms}
        if len(fs_bpms) == 1:
            return fs_bpms.pop()
        else:
            print('BPMs are not configured with the same ACQChannel.')
            return None

    def get_switching_frequency(self, rf_freq: float) -> float:
        """Return the switching frequency.

        Args:
            rf_freq (float): RF frequency.

        Returns:
            float: switching frequency.

        """
        fsw_bpms = {
            dev.get_switching_frequency(rf_freq) for dev in self.bpms}
        if len(fsw_bpms) == 1:
            return fsw_bpms.pop()
        else:
            print('BPMs are not configured with the same SwMode.')
            return None

    def mturn_config_acquisition(
            self, nr_points_after: int, nr_points_before=0,
            acq_rate='FAcq', repeat=True, external=True) -> int:
        """Configure acquisition for BPMs.

        Args:
            nr_points_after (int): number of points after trigger.
            nr_points_before (int): number of points after trigger.
                Defaults to 0.
            acq_rate (int|str, optional): Acquisition rate name ('TbT',
                'TbTPha', 'FOFB', 'FOFBPha', 'FAcq', 'ADC', 'ADCSwp') or value.
                Defaults to 'FAcq'.
            repeat (bool, optional): Whether or not acquisition should be
                repetitive. Defaults to True.
            external (bool, optional): Whether or not external trigger should
                be used. Defaults to True.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                <0: Index of the first BPM which did not stop last acq. plus 1.
                >0: Index of the first BPM which is not ready for acq. plus 1.

        """
        if acq_rate in self._csbpm.AcqChan:
            pass
        elif acq_rate.lower().startswith('facq'):
            acq_rate = self._csbpm.AcqChan.FAcq
        elif acq_rate.lower().startswith('fofbpha'):
            acq_rate = self._csbpm.AcqChan.FOFBPha
        elif acq_rate.lower().startswith('fofb'):
            acq_rate = self._csbpm.AcqChan.FOFB
        elif acq_rate.lower().startswith('tbtpha'):
            acq_rate = self._csbpm.AcqChan.TbTPha
        elif acq_rate.lower().startswith('tbt'):
            acq_rate = self._csbpm.AcqChan.TbT
        elif acq_rate.lower().startswith('adcswp'):
            acq_rate = self._csbpm.AcqChan.ADCSwp
        elif acq_rate.lower().startswith('adc'):
            acq_rate = self._csbpm.AcqChan.ADC
        else:
            raise ValueError(acq_rate + ' is not a valid acquisition rate.')

        if repeat:
            repeat = self._csbpm.AcqRepeat.Repetitive
        else:
            repeat = self._csbpm.AcqRepeat.Normal

        if external:
            trig = self._csbpm.AcqTrigTyp.External
        else:
            trig = self._csbpm.AcqTrigTyp.Now

        ret = self.cmd_mturn_acq_abort()
        if ret > 0:
            return -ret

        for bpm in self.bpms:
            bpm.acq_repeat = repeat
            bpm.acq_channel = acq_rate
            bpm.acq_trigger = trig
            bpm.acq_nrsamples_pre = nr_points_before
            bpm.acq_nrsamples_post = nr_points_after

        return self.cmd_mturn_acq_start()

    def cmd_mturn_acq_abort(self, wait=True, timeout=10) -> int:
        """Abort BPMs acquistion.

        Args:
            wait (bool, optional): whether or not to wait BPMs get ready.
                Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for bpm in self.bpms:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Abort

        if wait:
            return self.wait_acquisition_finish(timeout=timeout)
        return 0

    def wait_acquisition_finish(self, timeout=10) -> int:
        """Wait for all BPMs to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, bpm in enumerate(self.bpms):
            t0_ = _time.time()
            if not bpm.wait_acq_finish(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def cmd_mturn_acq_start(self, wait=True, timeout=10) -> int:
        """Start BPMs acquisition.

        Args:
            wait (bool, optional): whether or not to wait BPMs get ready.
                Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for bpm in self.bpms:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Start
        if wait:
            return self.wait_acquisition_start(timeout=timeout)
        return 0

    def wait_acquisition_start(self, timeout=10) -> bool:
        """Wait for all BPMs to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, bpm in enumerate(self.bpms):
            t0_ = _time.time()
            if not bpm.wait_acq_start(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def set_switching_mode(self, mode='direct'):
        """Set switching mode of BPMS.

        Args:
            mode ((str, int), optional): Desired mode, must be in
                {'direct', 'switching', 1, 3}. Defaults to 'direct'.

        Raises:
            ValueError: When mode is not in {'direct', 'switching', 1, 3}.

        """
        if mode not in ('direct', 'switching', 1, 3):
            raise ValueError('Value must be in ("direct", "switching", 1, 3).')

        for bpm in self.bpms:
            bpm.switching_mode = mode

    def mturn_update_initial_timestamps(self):
        """Call this method before acquisition to get orbit for comparison."""
        self._initial_timestamps = self.get_mturn_timestamps()

    def mturn_reset_flags(self):
        """Reset Multiturn flags to wait for a new orbit update."""
        for flag in self._mturn_flags.values():
            flag.clear()

    def mturn_reset_flags_and_update_initial_timestamps(self):
        """Set initial state to wait for orbit acquisition to start."""
        self.mturn_reset_flags()
        self.mturn_update_initial_timestamps()

    def mturn_wait_update_flags(self, timeout=10):
        """Wait for all acquisition flags to be updated.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, flag in enumerate(self._mturn_flags.values()):
            t00 = _time.time()
            if not flag.wait(timeout=timeout):
                return i + 1
            timeout -= _time.time() - t00
            timeout = max(timeout, 0)
        return 0

    def mturn_wait_update_timestamps(self, timeout=10) -> int:
        """Call this method after acquisition to check if data was updated.

        For this method to work it is necessary to call
            mturn_update_initial_timestamps
        before the acquisition starts, so that a reference for comparison is
        created.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int: code describing what happened:
                -2: size of timestamps changed in relation to initial timestamp
                -1: initial timestamps were not defined;
                =0: data updated.
                >0: index of the first BPM which did not update plus 1.

        """
        if self._initial_timestamps is None:
            return -1

        tsmp0 = self._initial_timestamps
        while timeout > 0:
            t00 = _time.time()
            tsmp = self.get_mturn_timestamps()
            if tsmp.size != tsmp0.size:
                return -2
            errors = _np.any(_np.equal(tsmp, tsmp0), axis=1)
            if not _np.any(errors):
                return 0
            _time.sleep(0.1)
            timeout -= _time.time() - t00

        return int(_np.nonzero(errors)[0][0])+1

    def mturn_wait_update(self, timeout=10) -> int:
        """Combine all methods to wait update data.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int: code describing what happened:
                -2: size of timestamps changed in relation to initial timestamp
                -1: initial timestamps were not defined;
                =0: data updated.
                >0: index of the first BPM which did not update plus 1.

        """
        t00 = _time.time()
        ret = self.mturn_wait_update_flags(timeout)
        if ret > 0:
            return ret
        timeout -= _time.time() - t00

        return self.mturn_wait_update_timestamps(timeout)

    def _mturn_set_flag(self, pvname, **kwargs):
        _ = kwargs
        self._mturn_flags[pvname].set()



"""Power Supply BSMP Constants.

Module for definitions of BSMP entities of power supply devices.

Documentation:

https://wiki-sirius.lnls.br/mediawiki/index.php/Machine:Power_Supplies
"""
from ...bsmp import constants as _const_bsmp

# version of the BSMP implementation of power supplies that is compatible
# with the current implemenation of this module.

__version__ = 'FRWM - V1.40 V0.44 - 30/06/2022'

# maximum number of power supplies controlled with a UDC
UDC_MAX_NR_DEV = 4

# --- Const DCDC ---

class ConstPSBSMP:
    """Namespace for organizing power supply BSMP constants."""

    # --- group of BSMP variables
    G_ALL = _const_bsmp.ID_STD_GROUP_ALL
    G_READONLY = _const_bsmp.ID_STD_GROUP_READONLY
    G_WRITE = _const_bsmp.ID_STD_GROUP_WRITE

    # --- types ---
    T_STATUS = 0
    T_STATE = 1
    T_REMOTE = 2
    T_MODEL = 3
    T_FLOAT = 4
    T_UINT8 = 5
    T_UINT16 = 6
    T_UINT32 = 7
    T_CHAR128 = 8
    T_FLOAT4 = 9
    T_PARAM = 10
    T_FLOAT12 = 11
    T_DSPCLASS = 12
    T_NONE = 13
    T_ENUM = 14

    # --- enums ---
    E_PS_MODEL_EMPTY = 0
    E_PS_MODEL_FBP = 1
    E_PS_MODEL_FBP_DCLink = 2
    E_PS_MODEL_FAC_ACDC = 3
    E_PS_MODEL_FAC_DCDC = 4
    E_PS_MODEL_FAC_2S_ACDC = 5
    E_PS_MODEL_FAC_2S_DCDC = 6
    E_PS_MODEL_FAC_2P4S_ACDC = 7
    E_PS_MODEL_FAC_2P4S_DCDC = 8
    E_PS_MODEL_FAP = 9
    E_PS_MODEL_FAP_4P = 10
    E_PS_MODEL_FAC_DCDC_EMA = 11
    E_PS_MODEL_FAP_2P2S = 12
    E_PS_MODEL_FAP_IMAS = 13
    E_PS_MODEL_FAC_2P_ACDC_IMAS = 14
    E_PS_MODEL_FAC_2P_DCDC_IMAS = 15
    E_PS_MODEL_UNINITIALIZED = 31

    E_STATE_OFF = 0
    E_STATE_INTERLOCK = 1
    E_STATE_INITIALIZING = 2
    E_STATE_SLOWREF = 3
    E_STATE_SLOWREFSYNC = 4
    E_STATE_CYCLE = 5
    E_STATE_RMPWFM = 6
    E_STATE_MIGWFM = 7   # deprecated?
    E_STATE_FASTREF = 8  # deprecated?

    E_INTERFACE_REMOTE = 0
    E_INTERFACE_LOCAL = 1
    E_INTERFACE_PCHOST = 2  # deprecated?

    E_WFMREFSYNC_SAMPLEBYSAMPLE = 0
    E_WFMREFSYNC_SAMPLEBYSAMPLE_ONECYCLE = 1
    E_WFMREFSYNC_ONESHOT = 2

    E_SIGGENTYPE_SINE = 0
    E_SIGGENTYPE_DAMPEDSINE = 1
    E_SIGGENTYPE_TRAPEZOIDAL = 2
    E_SIGGENTYPE_DAMPEDSQUAREDSINE = 3
    E_SIGGENTYPE_SQUARE = 4

    E_DSPCLASS_DSP_ERROR = 0
    E_DSPCLASS_DSP_SRLIM = 1
    E_DSPCLASS_DSP_LPF = 2
    E_DSPCLASS_DSP_PI = 3
    E_DSPCLASS_DSP_IIR_2P2Z = 4
    E_DSPCLASS_DSP_IIR_2P3Z = 5
    E_DSPCLASS_DSP_FF = 6
    E_DSPCLASS_DSP_VECT_PRODUCT = 7

    E_PARAMMEMORY_DEFAULT_INITIALIZATION = 1
    E_PARAMMEMORY_OFFBOARD_EEPROM = 2
    E_PARAMMEMORY_ONBOARD_EEPROM = 3

    # --- functions ---
    F_TURN_ON = 0
    F_TURN_OFF = 1
    F_OPEN_LOOP = 2
    F_CLOSE_LOOP = 3
    F_SELECT_OP_MODE = 4
    F_RESET_INTERLOCKS = 5
    F_SET_COMMAND_INTERFACE = 6
    F_SET_SERIAL_TERMINATION = 7
    F_UNLOCK_UDC = 8
    F_LOCK_UDC = 9
    F_CFG_SOURCE_SCOPE = 10
    F_CFG_FREQ_SCOPE = 11
    F_CFG_DURATION_SCOPE = 12
    F_ENABLE_SCOPE = 13
    F_DISABLE_SCOPE = 14
    F_SYNC_PULSE = 15
    F_SET_SLOWREF = 16
    F_SET_SLOWREF_FBP = 17
    F_SET_SLOWREF_READBACK_MON = 18
    F_SET_SLOWREF_FBP_READBACK_MON = 19
    F_SET_SLOWREF_READBACK_REF = 20
    F_SET_SLOWREF_FBP_READBACK_REF = 21
    F_RESET_COUNTERS = 22
    F_CFG_WFMREF = 23
    F_SELECT_WFMREF = 24
    F_GET_WFMREF_SIZE = 25
    F_RESET_WFMREF = 26
    F_CFG_SIGGEN = 27
    F_SET_SIGGEN = 28
    F_ENABLE_SIGGEN = 29
    F_DISABLE_SIGGEN = 30
    F_SET_PARAM = 31
    F_GET_PARAM = 32
    F_SAVE_PARAM_EEPROM = 33
    F_LOAD_PARAM_EEPROM = 34
    F_SAVE_PARAM_BANK = 35
    F_LOAD_PARAM_BANK = 36
    F_SET_DSP_COEFFS = 37
    F_GET_DSP_COEFF = 38
    F_SAVE_DSP_COEFFS_EEPROM = 39
    F_LOAD_DSP_COEFFS_EEPROM = 40
    F_SAVE_DSP_MODULES_EEPROM = 41
    F_LOAD_DSP_MODULES_EEPROM = 42
    F_RESET_UDC = 43

    # --- common variables ---
    V_PS_STATUS = 0
    V_PS_SETPOINT = 1  # corresponds to IOC Current-RB
    V_PS_REFERENCE = 2  # corresponds to IOC CurrentRef-Mon
    V_FIRMWARE_VERSION = 3
    V_COUNTER_SET_SLOWREF = 4
    V_COUNTER_SYNC_PULSE = 5
    V_SIGGEN_ENABLE = 6
    V_SIGGEN_TYPE = 7
    V_SIGGEN_NUM_CYCLES = 8
    V_SIGGEN_N = 9
    V_SIGGEN_FREQ = 10
    V_SIGGEN_AMPLITUDE = 11
    V_SIGGEN_OFFSET = 12
    V_SIGGEN_AUX_PARAM = 13
    V_WFMREF_SELECTED = 14
    V_WFMREF_SYNC_MODE = 15
    V_WFMREF_FREQUENCY = 16
    V_WFMREF_GAIN = 17
    V_WFMREF_OFFSET = 18
    V_WFMREF0_START = 19
    V_WFMREF0_END = 20
    V_WFMREF0_IDX = 21
    V_WFMREF1_START = 22
    V_WFMREF1_END = 23
    V_WFMREF1_IDX = 24
    V_SCOPE_FREQUENCY = 25
    V_SCOPE_DURATION = 26
    V_SCOPE_SRC_DATA = 27
    V_PERIOD_SYNC_PULSE = 28

    # --- undefined variables
    V_UNDEF29 = 29
    V_UNDEF30 = 30

    # --- power supply parameters ---
    # ----- class PS -----
    P_PS_NAME = 0
    P_PS_MODEL = 1
    P_PS_NR_PSMODELS = 2
    # ----- class Communication -----
    P_COMM_CMD_INTERFACE = 3
    P_COMM_RS485_BAUDRATE = 4
    P_COMM_RS485_ADDRESS = 5
    P_COMM_RS485_TERMINATOR_RESISTOR = 6
    P_COMM_UDC_NETWORK_ADDRESS = 7
    P_COMM_ETHERNET_IP = 8
    P_COMM_ETHERNET_SUBNET_MASK = 9
    P_COMM_BUZZER_VOLUME = 10
    # ----- class Control -----
    P_CTRL_FREQ_CONTROL_ISR = 11
    P_CTRL_FREQ_TIME_SLICER = 12
    P_CTRL_CONTROL_LOOP_STATE = 13
    P_CTRL_MAX_REF = 14
    P_CTRL_MIN_REF = 15
    P_CTRL_MAX_REF_OPEN_LOOP = 16
    P_CTRL_MIN_REF_OPEN_LOOP = 17
    # ----- class PWM -----
    P_PWM_FREQ = 18
    P_PWM_DEAD_TIME = 19
    P_PWM_MAX_DUTY = 20
    P_PWM_MIN_DUTY = 21
    P_PWM_MAX_DUTY_OPEN_LOOP = 22
    P_PWM_MIN_DUTY_OPEN_LOOP = 23
    P_PWM_LIM_DUTY_SHARE = 24
    # ----- class HRADC -----
    P_HRADC_NR_BOARDS = 25
    P_HRADC_SPI_CLK = 26
    P_HRADC_FREQ_SAMPLING = 27
    P_HRADC_ENABLE_HEATER = 28
    P_HRADC_ENABLE_RAILS_MON = 29
    P_HRADC_TRANSDUCER_OUTPUT = 30
    P_HRADC_TRANSDUCER_GAIN = 31
    P_HRADC_TRANSDUCER_OFFSET = 32
    # ----- class SigGen -----
    P_SIGGEN_TYPE = 33
    P_SIGGEN_NUM_CYCLES = 34
    P_SIGGEN_FREQ = 35
    P_SIGGEN_AMPLITUDE = 36
    P_SIGGEN_OFFSET = 37
    P_SIGGEN_AUX_PARAM = 38
    # ----- class WfmRef -----
    P_WFMREF_ID = 39
    P_WFMREF_SYNC_MODE = 40
    P_WFMREF_FREQ = 41
    P_WFMREF_GAIN = 42
    P_WFMREF_OFFSET = 43
    # ----- class Analog Variables -----
    P_ANALOG_MAX = 44
    P_ANALOG_MIN = 45
    # ----- class Debounding manager -----
    P_HARD_INTLK_DEBOUNCE_TIME = 46
    P_HARD_INTLK_RESET_TIME = 47
    P_SOFT_INTLK_DEBOUNCE_TIME = 48
    P_SOFT_INTLK_RESET_TIME = 49
    # ---- Scope -----
    P_SCOPE_SAMPLING_FREQUENCY = 50
    P_SCOPE_DATA_SOURCE = 51

    # --- acknowledgement return ---
    ACK_OK = 0
    ACK_PS_LOCAL_STATE = 1
    ACK_PS_PCHOST_STATE = 2
    ACK_PS_INTERLOCK = 3
    ACK_UDC_LOCKED = 4
    ACK_DSP_TIMEOUT = 5
    ACK_DSP_BUSY = 6
    ACK_RESOURCE_BUSY = 7
    ACK_INVALID_CMD = 8


class ConstFBP(ConstPSBSMP):
    """Namespace for organizing power supply FBP BSMP constants."""

    # --- group of BSMP variables
    G_SOFB = 3

    # --- FSB variables ---
    V_PS_SOFT_INTERLOCKS = 31  # BSMP doc says ID numb. should be continous!
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD = 33  # corresponds to IOC Current-Mon
    V_V_LOAD = 34
    V_V_DCLINK = 35
    V_TEMP_SWITCHES = 36
    V_DUTY_CYCLE = 37  # (float)
    V_PS_ALARMS = 38
    # --- FBP mirror variables ----
    V_PS_STATUS1 = 46
    V_PS_STATUS2 = 47
    V_PS_STATUS3 = 48
    V_PS_STATUS4 = 49
    V_PS_SETPOINT1 = 50
    V_PS_SETPOINT2 = 51
    V_PS_SETPOINT3 = 52
    V_PS_SETPOINT4 = 53
    V_PS_REFERENCE1 = 54
    V_PS_REFERENCE2 = 55
    V_PS_REFERENCE3 = 56
    V_PS_REFERENCE4 = 57
    V_PS_SOFT_INTERLOCK1 = 58
    V_PS_SOFT_INTERLOCK2 = 59
    V_PS_SOFT_INTERLOCK3 = 60
    V_PS_SOFT_INTERLOCK4 = 61
    V_PS_HARD_INTERLOCK1 = 62
    V_PS_HARD_INTERLOCK2 = 63
    V_PS_HARD_INTERLOCK3 = 64
    V_PS_HARD_INTERLOCK4 = 65
    V_I_LOAD1 = 66
    V_I_LOAD2 = 67
    V_I_LOAD3 = 68
    V_I_LOAD4 = 69
    V_PS_ALARMS1 = 70
    V_PS_ALARMS2 = 71
    V_PS_ALARMS3 = 72
    V_PS_ALARMS4 = 73


class ConstFAC_DCDC(ConstPSBSMP):
    """Namespace for organizing power supply FAC_DCDC BSMP constants."""

    # --- FAC_DCDC variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD_MEAN = 33
    V_I_LOAD1 = 34
    V_I_LOAD2 = 35
    V_V_CAPBANK = 36
    V_DUTY_CYCLE = 37
    V_I_LEAKAGE = 38
    V_V_INPUT_IIB = 39
    V_I_INPUT_IIB = 40
    V_I_OUTPUT_IIB = 41
    V_TEMP_IGBT_1_IIB = 42
    V_TEMP_IGBT_2_IIB = 43
    V_TEMP_INDUCTOR_IIB = 44
    V_TEMP_HEATSINK_IIB = 45
    V_V_DRIVER_IIB = 46
    V_I_DRIVER_1_IIB = 47
    V_I_DRIVER_2_IIB = 48
    V_TEMP_BOARD_IIB = 49
    V_RH_IIB = 50
    V_IIB_INTERLOCKS = 51
    V_IIB_ALARMS = 52
    V_PS_ALARMS = 53


class ConstFAC_2P4S_DCDC(ConstPSBSMP):
    """Namespace for organizing power supply FAC_2P4S_DCDC BSMP constants."""

    # --- FAC_2P4S_DCDC variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD_MEAN = 33
    V_I_LOAD1 = 34
    V_I_LOAD2 = 35
    V_I_ARM_1 = 36
    V_I_ARM_2 = 37
    V_V_CAPBANK_1 = 38
    V_V_CAPBANK_2 = 39
    V_V_CAPBANK_3 = 40
    V_V_CAPBANK_4 = 41
    V_V_CAPBANK_5 = 42
    V_V_CAPBANK_6 = 43
    V_V_CAPBANK_7 = 44
    V_V_CAPBANK_8 = 45
    V_DUTY_CYCLE_1 = 46
    V_DUTY_CYCLE_2 = 47
    V_DUTY_CYCLE_3 = 48
    V_DUTY_CYCLE_4 = 49
    V_DUTY_CYCLE_5 = 50
    V_DUTY_CYCLE_6 = 51
    V_DUTY_CYCLE_7 = 52
    V_DUTY_CYCLE_8 = 53
    V_V_INPUT_IIB_A = 54
    V_I_INPUT_IIB_A = 55
    V_I_OUTPUT_IIB_A = 56
    V_TEMP_IGBTS_1_IIB_A = 57
    V_TEMP_IGBTS_2_IIB_A = 58
    V_TEMP_INDUCTOR_IIB_A = 59
    V_TEMP_HEATSINK_IIB_A = 60
    V_V_DRIVER_IIB_A = 61
    V_I_DRIVER_1_IIB_A = 62
    V_I_DRIVER_2_IIB_A = 63
    V_TEMP_BOARD_IIB_A = 64
    V_RH_IIB_A = 65
    V_IIB_INTERLOCKS_A = 66
    V_IIB_ALARMS_A = 67
    V_V_INPUT_IIB_B = 68
    V_I_INPUT_IIB_B = 69
    V_I_OUTPUT_IIB_B = 70
    V_TEMP_IGBTS_1_IIB_B = 71
    V_TEMP_IGBTS_2_IIB_B = 72
    V_TEMP_INDUCTOR_IIB_B = 73
    V_TEMP_HEATSINK_IIB_B = 74
    V_V_DRIVER_IIB_B = 75
    V_I_DRIVER_1_IIB_B = 76
    V_I_DRIVER_2_IIB_B = 77
    V_TEMP_BOARD_IIB_B = 78
    V_RH_IIB_B = 79
    V_IIB_INTERLOCKS_B = 80
    V_IIB_ALARMS_B = 81
    V_PS_ALARMS = 82


class ConstFAC_2S_DCDC(ConstPSBSMP):
    """Namespace for organizing power supply FAC_2S_DCDC BSMP constants."""

    # --- FAC_2S_DCDC variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD_MEAN = 33
    V_I_LOAD1 = 34
    V_I_LOAD2 = 35
    V_V_CAPBANK_1 = 36
    V_V_CAPBANK_2 = 37
    V_DUTY_CYCLE_1 = 38
    V_DUTY_CYCLE_2 = 39
    V_V_INPUT_IIB_1 = 40
    V_I_INPUT_IIB_1 = 41
    V_I_OUTPUT_IIB_1 = 42
    V_TEMP_IGBTS_1_IIB_1 = 43
    V_TEMP_IGBTS_2_IIB_1 = 44
    V_TEMP_INDUCTOR_IIB_1 = 45
    V_TEMP_HEATSINK_IIB_1 = 46
    V_V_DRIVER_IIB_1 = 47
    V_I_DRIVER_1_IIB_1 = 48
    V_I_DRIVER_2_IIB_1 = 49
    V_TEMP_BOARD_IIB_1 = 50
    V_RH_IIB_1 = 51
    V_IIB_INTERLOCKS_1 = 52
    V_IIB_ALARMS_1 = 53
    V_V_INPUT_IIB_2 = 54
    V_I_INPUT_IIB_2 = 55
    V_I_OUTPUT_IIB_2 = 56
    V_TEMP_IGBTS_1_IIB_2 = 57
    V_TEMP_IGBTS_2_IIB_2 = 58
    V_TEMP_INDUCTOR_IIB_2 = 59
    V_TEMP_HEATSINK_IIB_2 = 60
    V_V_DRIVER_IIB_2 = 61
    V_I_DRIVER_1_IIB_2 = 62
    V_I_DRIVER_2_IIB_2 = 63
    V_TEMP_BOARD_IIB_2 = 64
    V_RH_IIB_2 = 65
    V_IIB_INTERLOCKS_2 = 66
    V_IIB_ALARMS_2 = 67
    V_PS_ALARMS = 68


class ConstFAP(ConstPSBSMP):
    """Namespace for organizing power supply FAP BSMP constants."""

    # --- FAP variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD_MEAN = 33  # corresponds to IOC Current-Mon
    V_I_LOAD1 = 34  # corresponds to IOC Current1-Mon
    V_I_LOAD2 = 35  # corresponds to IOC Current2-Mon
    V_V_DCLINK = 36
    V_I_IGBT_1 = 37
    V_I_IGBT_2 = 38
    V_DUTY_CYCLE_1 = 39
    V_DUTY_CYCLE_2 = 40
    V_DUTY_DIFF = 41
    V_V_INPUT_IIB = 42
    V_V_OUTPUT_IIB = 43
    V_I_IGBT_1_IIB = 44
    V_I_IGBT_2_IIB = 45
    V_TEMP_IGBT_1_IIB = 46
    V_TEMP_IGBT_2_IIB = 47
    V_V_DRIVER_IIB = 48
    V_I_DRIVER_1_IIB = 49
    V_I_DRIVER_2_IIB = 50
    V_TEMP_INDUCTOR_IIB = 51
    V_TEMP_HEATSINK_IIB = 52
    V_I_LEAKAGE_IIB = 53
    V_TEMP_BOARD_IIB = 54
    V_RH_IIB = 55
    V_IIB_INTERLOCKS = 56
    V_IIB_ALARMS = 57
    V_PS_ALARMS = 58


class ConstFAP_4P(ConstPSBSMP):
    """Namespace for organizing power supply FAP_4P BSMP constants."""

    # --- FAP_4P variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD_MEAN = 33
    V_I_LOAD1 = 34
    V_I_LOAD2 = 35
    V_V_LOAD = 36
    V_I_IGBT_1_1 = 37
    V_I_IGBT_2_1 = 38
    V_I_IGBT_1_2 = 39
    V_I_IGBT_2_2 = 40
    V_I_IGBT_1_3 = 41
    V_I_IGBT_2_3 = 42
    V_I_IGBT_1_4 = 43
    V_I_IGBT_2_4 = 44
    V_V_DCLINK_1 = 45
    V_V_DCLINK_2 = 46
    V_V_DCLINK_3 = 47
    V_V_DCLINK_4 = 48
    V_DUTY_MEAN = 49
    V_DUTY_CYCLE_1_1 = 50
    V_DUTY_CYCLE_2_1 = 51
    V_DUTY_CYCLE_1_2 = 52
    V_DUTY_CYCLE_2_2 = 53
    V_DUTY_CYCLE_1_3 = 54
    V_DUTY_CYCLE_2_3 = 55
    V_DUTY_CYCLE_1_4 = 56
    V_DUTY_CYCLE_2_4 = 57
    V_I_LEAKAGE = 58
    V_V_INPUT_IIB_1 = 59
    V_V_OUTPUT_IIB_1 = 60
    V_I_IGBT_1_IIB_1 = 61
    V_I_IGBT_2_IIB_1 = 62
    V_TEMP_IGBT_1_IIB_1 = 63
    V_TEMP_IGBT_2_IIB_1 = 64
    V_V_DRIVER_IIB_1 = 65
    V_I_DRIVER_1_IIB_1 = 66
    V_I_DRIVER_2_IIB_1 = 67
    V_TEMP_INDUCTOR_IIB_1 = 68
    V_TEMP_HEATSINK_IIB_1 = 69
    V_TEMP_BOARD_IIB_1 = 70
    V_RH_IIB_1 = 71
    V_IIB_INTERLOCKS_1 = 72
    V_IIB_ALARMS_1 = 73
    V_V_INPUT_IIB_2 = 74
    V_V_OUTPUT_IIB_2 = 75
    V_I_IGBT_1_IIB_2 = 76
    V_I_IGBT_2_IIB_2 = 77
    V_TEMP_IGBT_1_IIB_2 = 78
    V_TEMP_IGBT_2_IIB_2 = 79
    V_V_DRIVER_IIB_2 = 80
    V_I_DRIVER_1_IIB_2 = 81
    V_I_DRIVER_2_IIB_2 = 82
    V_TEMP_INDUCTOR_IIB_2 = 83
    V_TEMP_HEATSINK_IIB_2 = 84
    V_TEMP_BOARD_IIB_2 = 85
    V_RH_IIB_2 = 86
    V_IIB_INTERLOCKS_2 = 87
    V_IIB_ALARMS_2 = 88
    V_V_INPUT_IIB_3 = 89
    V_V_OUTPUT_IIB_3 = 90
    V_I_IGBT_1_IIB_3 = 91
    V_I_IGBT_2_IIB_3 = 92
    V_TEMP_IGBT_1_IIB_3 = 93
    V_TEMP_IGBT_2_IIB_3 = 94
    V_V_DRIVER_IIB_3 = 95
    V_I_DRIVER_1_IIB_3 = 96
    V_I_DRIVER_2_IIB_3 = 97
    V_TEMP_INDUCTOR_IIB_3 = 98
    V_TEMP_HEATSINK_IIB_3 = 99
    V_TEMP_BOARD_IIB_3 = 100
    V_RH_IIB_3 = 101
    V_IIB_INTERLOCKS_3 = 102
    V_IIB_ALARMS_3 = 103
    V_V_INPUT_IIB_4 = 104
    V_V_OUTPUT_IIB_4 = 105
    V_I_IGBT_1_IIB_4 = 106
    V_I_IGBT_2_IIB_4 = 107
    V_TEMP_IGBT_1_IIB_4 = 108
    V_TEMP_IGBT_2_IIB_4 = 109
    V_V_DRIVER_IIB_4 = 110
    V_I_DRIVER_1_IIB_4 = 111
    V_I_DRIVER_2_IIB_4 = 112
    V_TEMP_INDUCTOR_IIB_4 = 113
    V_TEMP_HEATSINK_IIB_4 = 114
    V_TEMP_BOARD_IIB_4 = 115
    V_RH_IIB_4 = 116
    V_IIB_INTERLOCKS_4 = 117
    V_IIB_ALARMS_4 = 118
    V_PS_ALARMS = 119


class ConstFAP_2P2S(ConstPSBSMP):
    """Namespace for organizing power supply FAP_2P2S BSMP constants."""

    # --- FAP_2P2S variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_I_LOAD_MEAN = 33
    V_I_LOAD1 = 34
    V_I_LOAD2 = 35
    V_I_ARM_1 = 36
    V_I_ARM_2 = 37
    V_I_IGBT_1_1 = 38
    V_I_IGBT_2_1 = 39
    V_I_IGBT_1_2 = 40
    V_I_IGBT_2_2 = 41
    V_I_IGBT_1_3 = 42
    V_I_IGBT_2_3 = 43
    V_I_IGBT_1_4 = 44
    V_I_IGBT_2_4 = 45
    V_I_MOD_1 = 46
    V_I_MOD_2 = 47
    V_I_MOD_3 = 48
    V_I_MOD_4 = 49
    V_V_DCLINK_1 = 50
    V_V_DCLINK_2 = 51
    V_V_DCLINK_3 = 52
    V_V_DCLINK_4 = 53
    V_DUTY_MEAN = 54
    V_DUTY_ARMS_DIFF = 55
    V_DUTY_CYCLE_1_1 = 56
    V_DUTY_CYCLE_2_1 = 57
    V_DUTY_CYCLE_1_2 = 58
    V_DUTY_CYCLE_2_2 = 59
    V_DUTY_CYCLE_1_3 = 60
    V_DUTY_CYCLE_2_3 = 61
    V_DUTY_CYCLE_1_4 = 62
    V_DUTY_CYCLE_2_4 = 63
    V_I_LEAKAGE = 64
    V_V_INPUT_IIB_1 = 65
    V_V_OUTPUT_IIB_1 = 66
    V_I_IGBT_1_IIB_1 = 67
    V_I_IGBT_2_IIB_1 = 68
    V_TEMP_IGBT_1_IIB_1 = 69
    V_TEMP_IGBT_2_IIB_1 = 70
    V_V_DRIVER_IIB_1 = 71
    V_I_DRIVER_1_IIB_1 = 72
    V_I_DRIVER_2_IIB_1 = 73
    V_TEMP_INDUCTOR_IIB_1 = 74
    V_TEMP_HEATSINK_IIB_1 = 75
    V_TEMP_BOARD_IIB_1 = 76
    V_RH_IIB_1 = 77
    V_IIB_INTERLOCKS_1 = 78
    V_IIB_ALARMS_1 = 79
    V_V_INPUT_IIB_2 = 80
    V_V_OUTPUT_IIB_2 = 81
    V_I_IGBT_1_IIB_2 = 82
    V_I_IGBT_2_IIB_2 = 83
    V_TEMP_IGBT_1_IIB_2 = 84
    V_TEMP_IGBT_2_IIB_2 = 85
    V_V_DRIVER_IIB_2 = 86
    V_I_DRIVER_1_IIB_2 = 87
    V_I_DRIVER_2_IIB_2 = 88
    V_TEMP_INDUCTOR_IIB_2 = 89
    V_TEMP_HEATSINK_IIB_2 = 90
    V_TEMP_BOARD_IIB_2 = 91
    V_RH_IIB_2 = 92
    V_IIB_INTERLOCKS_2 = 93
    V_IIB_ALARMS_2 = 94
    V_V_INPUT_IIB_3 = 95
    V_V_OUTPUT_IIB_3 = 96
    V_I_IGBT_1_IIB_3 = 97
    V_I_IGBT_2_IIB_3 = 98
    V_TEMP_IGBT_1_IIB_3 = 99
    V_TEMP_IGBT_2_IIB_3 = 100
    V_V_DRIVER_IIB_3 = 101
    V_I_DRIVER_1_IIB_3 = 102
    V_I_DRIVER_2_IIB_3 = 103
    V_TEMP_INDUCTOR_IIB_3 = 104
    V_TEMP_HEATSINK_IIB_3 = 105
    V_TEMP_BOARD_IIB_3 = 106
    V_RH_IIB_3 = 107
    V_IIB_INTERLOCKS_3 = 108
    V_IIB_ALARMS_3 = 109
    V_V_INPUT_IIB_4 = 110
    V_V_OUTPUT_IIB_4 = 111
    V_I_IGBT_1_IIB_4 = 112
    V_I_IGBT_2_IIB_4 = 113
    V_TEMP_IGBT_1_IIB_4 = 114
    V_TEMP_IGBT_2_IIB_4 = 115
    V_V_DRIVER_IIB_4 = 116
    V_I_DRIVER_1_IIB_4 = 117
    V_I_DRIVER_2_IIB_4 = 118
    V_TEMP_INDUCTOR_IIB_4 = 119
    V_TEMP_HEATSINK_IIB_4 = 120
    V_TEMP_BOARD_IIB_4 = 121
    V_RH_IIB_4 = 122
    V_IIB_INTERLOCKS_4 = 123
    V_IIB_ALARMS_4 = 124
    V_PS_ALARMS = 125


# --- Const ACDC ---

class ConstFBP_DCLink(ConstPSBSMP):
    """Namespace for organizing power supply FBP_DCLink BSMP constants."""

    # --- FBP_DCLink variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_MODULES_STATUS = 33
    V_V_OUT = 34
    V_V_OUT_1 = 35
    V_V_OUT_2 = 36
    V_V_OUT_3 = 37
    V_DIG_POT_TAP = 38


class ConstFAC_2S_ACDC(ConstPSBSMP):
    """Namespace for organizing power supply FAC_2S_ACDC BSMP constants."""

    # --- FAC_2S_ACDC variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_V_CAPBANK = 33
    V_I_OUT_RECTIFIER = 34
    V_DUTY_CYCLE = 35
    V_I_INPUT_IS_IIB = 36
    V_V_INPUT_IS_IIB = 37
    V_TEMP_IGBT_IS_IIB = 38
    V_V_DRIVER_IS_IIB = 39
    V_I_DRIVER_IS_IIB = 40
    V_TEMP_INDUCTOR_IS_IIB = 41
    V_TEMP_HEATSINK_IS_IIB = 42
    V_TEMP_BOARD_IS_IIB = 43
    V_RH_IS_IIB = 44
    V_IIB_INTERLOCKS_IS = 45
    V_IIB_ALARMS_IS = 46
    V_V_OUTPUT_CMD_IIB = 47
    V_V_CAPBANK_CMD_IIB = 48
    V_TEMP_RECT_INDUCTOR_CMD_IIB = 49
    V_TEMP_RECT_HEATSINK_CMD_IIB = 50
    V_V_EXT_BOARDS_CMD_IIB = 51
    V_I_AUX_BOARD_CMD_IIB = 52
    V_I_IDB_BOARD_CMD_IIB = 53
    V_I_LEAKAGE_CMD_IIB = 54
    V_TEMP_BOARD_CMD_IIB = 55
    V_RH_CMD_IIB = 56
    V_IIB_INTERLOCKS_CMD = 57
    V_IIB_ALARMS_CMD = 58


class ConstFAC_2P4S_ACDC(ConstPSBSMP):
    """Namespace for organizing power supply FAC_2P4S_ACDC BSMP constants."""

    # --- FAC_2P4S_ACDC variables ---
    V_PS_SOFT_INTERLOCKS = 31
    V_PS_HARD_INTERLOCKS = 32
    V_V_CAPACITOR_BANK = 33
    V_I_OUT_RECTIFIER = 34
    V_DUTY_CYCLE = 35
    V_I_INPUT_IS_IIB = 36
    V_V_INPUT_IS_IIB = 37
    V_TEMP_IGBT_IS_IIB = 38
    V_V_DRIVER_IS_IIB = 39
    V_I_DRIVER_IS_IIB = 40
    V_TEMP_INDUCTOR_IS_IIB = 41
    V_TEMP_HEATSINK_IS_IIB = 42
    V_TEMP_BOARD_IS_IIB = 43
    V_RH_IS_IIB = 44
    V_IIB_INTERLOCKS_IS = 45
    V_IIB_ALARMS_IS = 46
    V_V_OUTPUT_CMD_IIB = 47
    V_V_CAPBANK_CMD_IIB = 48
    V_TEMP_RECT_INDUCTOR_CMD_IIB = 49
    V_TEMP_RECT_HEATSINK_CMD_IIB = 50
    V_V_EXT_BOARDS_CMD_IIB = 51
    V_I_AUX_BOARD_CMD_IIB = 52
    V_I_IDB_BOARD_CMD_IIB = 53
    V_I_LEAKAGE_CMD_IIB = 54
    V_TEMP_BOARD_CMD_IIB = 55
    V_RH_CMD_IIB = 56
    V_IIB_INTERLOCKS_CMD = 57
    V_IIB_ALARMS_CMD = 58



"""High Level FOFB main application."""

import os as _os
import logging as _log
import time as _time
from functools import partial as _part
import epics as _epics
import numpy as _np

from ..util import update_bit as _updt_bit, get_bit as _get_bit
from ..epics import PV as _PV
from ..callbacks import Callback as _Callback
from ..envars import VACA_PREFIX as _vaca_prefix
from ..namesys import SiriusPVName as _PVName
from ..devices import FamFOFBControllers as _FamFOFBCtrls, Device as _Device,\
    FamFastCorrs as _FamFastCorrs, SOFB as _SOFB, RFGen as _RFGen

from .csdev import HLFOFBConst as _Const, ETypes as _ETypes


class App(_Callback):
    """High Level FOFB main application."""

    SCAN_FREQUENCY = 1  # [Hz]

    def __init__(self, tests=False):
        """Class constructor."""
        super().__init__()
        self._const = _Const()
        self._pvs_database = self._const.get_hlfofb_database()
        self._tests = tests
        self._init = False

        # internal states
        self._loop_state = self._const.LoopState.Open
        self._loop_state_lastsp = self._const.LoopState.Open
        self._loop_gain_h = 0.0520
        self._loop_gain_mon_h = 0
        self._loop_gain_v = 0.0520
        self._loop_gain_mon_v = 0
        self._thread_loopstate = None
        self._abort_thread = False
        self._loop_max_orb_dist = self._const.DEF_MAX_ORB_DISTORTION
        self._loop_max_orb_dist_enbl = self._const.DsblEnbl.Dsbl
        self._loop_packloss_detec_enbl = self._const.DsblEnbl.Dsbl
        self._corr_status = self._pvs_database['CorrStatus-Mon']['value']
        self._corr_setcurrzero_dur = 5
        self._thread_currzero = None
        self._abort_thread_currzero = False
        self._ch_maxacccurr = self._pvs_database['CHAccSatMax-RB']['value']
        self._cv_maxacccurr = self._pvs_database['CVAccSatMax-RB']['value']
        self._time_frame_len = self._pvs_database['TimeFrameLen-RB']['value']
        self._fofbctrl_status = self._pvs_database['CtrlrStatus-Mon']['value']
        self._thread_syncnet = None
        self._thread_reset = None
        self._fofbctrl_syncenbllist = _np.ones(self._const.nr_bpms, dtype=bool)
        self._fofbctrl_syncuseenbllist = self._const.DsblEnbl.Enbl
        self._reforb_x = _np.zeros(self._const.nr_bpms, dtype=float)
        self._reforbhw_x = _np.zeros(self._const.nr_bpms, dtype=float)
        self._reforb_y = _np.zeros(self._const.nr_bpms, dtype=float)
        self._reforbhw_y = _np.zeros(self._const.nr_bpms, dtype=float)
        self._respmat = _np.zeros(
            [2*self._const.nr_bpms, self._const.nr_corrs], dtype=float)
        self._invrespmat = self._respmat.copy().T
        self._invrespmatconv = self._invrespmat.copy()
        self._enable_lists = {
            'bpmx': _np.ones(self._const.nr_bpms, dtype=bool),
            'bpmy': _np.ones(self._const.nr_bpms, dtype=bool),
            'ch': _np.ones(self._const.nr_ch, dtype=bool),
            'cv': _np.ones(self._const.nr_cv, dtype=bool),
            'rf': _np.ones(1, dtype=bool),
        }
        self._corr_accdec_val = 1
        self._corr_accdec_enm = self._const.DecOpt.FOFB
        self._thread_enbllist = None
        self._abort_thread_enbllist = False
        self._min_sing_val = self._const.MIN_SING_VAL
        self._tikhonov_reg_const = self._const.TIKHONOV_REG_CONST
        self._invrespmat_normmode = self._const.GlobIndiv.Global
        self._pscoeffs = self._invrespmatconv[:-1].copy()
        self._psgains = _np.ones(self._const.nr_chcv, dtype=float)
        self._meas_respmat_kick = {
            'ch': 15,  # [urad]
            'cv': 22.5,  # [urad]
            'rf': 75,  # [Hz]
        }
        self._meas_respmat_wait = 1  # [s]
        self._meas_respmat_thread = None
        self._measuring_respmat = False

        # devices and connections
        self._sisofb_dev = _SOFB(_SOFB.DEVICES.SI)

        corrnames = self._const.ch_names + self._const.cv_names
        self._corrs_dev = _FamFastCorrs(corrnames)

        self._rf_dev = _RFGen()

        self._llfofb_dev = _FamFOFBCtrls()

        self._intlk_pvs = list()
        self._intlk_values = dict()
        for dev in self._llfofb_dev.ctrlrefdevs.values():
            pvo = dev.pv_object('LoopIntlk-Mon')
            self._intlk_values[pvo.pvname] = 0
            pvo.add_callback(
                self._callback_loopintlk, with_ctrlvars=False)
            self._intlk_pvs.append(pvo)

        self._corrs_dev.wait_for_connection(self._const.DEF_TIMEWAIT)

        self._auxbpm = _Device(
            'SI-01M1:DI-BPM',
            props2init=('INFOFOFBRate-RB', 'INFOMONITRate-RB'))

        havebeam_pvname = _PVName(
            'SI-Glob:AP-CurrInfo:StoredEBeam-Mon').substitute(
                prefix=_vaca_prefix)
        self._havebeam_pv = _PV(
            havebeam_pvname, connection_timeout=0.05,
            callback=self._callback_havebeam)

        # pvs to write methods
        self.map_pv2write = {
            'LoopState-Sel': self.set_loop_state,
            'LoopGainH-SP': _part(self.set_loop_gain, 'h'),
            'LoopGainV-SP': _part(self.set_loop_gain, 'v'),
            'LoopMaxOrbDistortion-SP': self.set_loop_max_orbit_dist,
            'LoopMaxOrbDistortionEnbl-Sel': self.set_loop_max_orbit_dist_enbl,
            'LoopPacketLossDetecEnbl-Sel': self.set_loop_packloss_detec_enbl,
            'CorrConfig-Cmd': self.cmd_corr_configure,
            'CorrSetPwrStateOn-Cmd': self.cmd_corr_pwrstate_on,
            'CorrSetPwrStateOff-Cmd': self.cmd_corr_pwrstate_off,
            'CorrSetOpModeManual-Cmd': self.cmd_corr_opmode_manual,
            'CorrSetAccFreezeDsbl-Cmd': self.cmd_corr_accfreeze_dsbl,
            'CorrSetAccFreezeEnbl-Cmd': self.cmd_corr_accfreeze_enbl,
            'CorrSetAccClear-Cmd': self.cmd_corr_accclear,
            'CorrSetCurrZero-Cmd': self.cmd_corr_currzero,
            'CorrSetCurrZeroDuration-SP': self.set_corr_currzero_duration,
            'CHAccSatMax-SP': _part(self.set_corr_accsatmax, 'ch'),
            'CVAccSatMax-SP': _part(self.set_corr_accsatmax, 'cv'),
            'TimeFrameLen-SP': self.set_timeframelen,
            'CtrlrConfBPMId-Cmd': self.cmd_fofbctrl_confbpmid,
            'CtrlrSyncNet-Cmd': self.cmd_fofbctrl_syncnet,
            'CtrlrSyncRefOrb-Cmd': self.cmd_fofbctrl_syncreforb,
            'CtrlrSyncTFrameLen-Cmd': self.cmd_fofbctrl_synctframelen,
            'CtrlrConfBPMLogTrg-Cmd': self.cmd_fofbctrl_confbpmlogtrg,
            'CtrlrSyncUseEnblList-Sel': self.set_fofbctrl_syncuseenablelist,
            'CtrlrSyncMaxOrbDist-Cmd': self.cmd_fofbctrl_syncmaxorbdist,
            'CtrlrSyncPacketLossDetec-Cmd': self.cmd_fofbctrl_syncpacklossdet,
            'CtrlrReset-Cmd': self.cmd_fofbctrl_reset,
            'CtrlrDsblSYSIDExc-Cmd': self.cmd_fofbctrl_dsblsysid,
            'FOFBAccDecimation-Sel': _part(self.set_corr_accdec, 'enum'),
            'FOFBAccDecimation-SP': _part(self.set_corr_accdec, 'value'),
            'RefOrbX-SP': _part(self.set_reforbit, 'x'),
            'RefOrbY-SP': _part(self.set_reforbit, 'y'),
            'RespMat-SP': self.set_respmat,
            'BPMXEnblList-SP': _part(self.set_enbllist, 'bpmx'),
            'BPMYEnblList-SP': _part(self.set_enbllist, 'bpmy'),
            'CHEnblList-SP': _part(self.set_enbllist, 'ch'),
            'CVEnblList-SP': _part(self.set_enbllist, 'cv'),
            'UseRF-Sel': _part(self.set_enbllist, 'rf'),
            'MinSingValue-SP': self.set_min_sing_value,
            'TikhonovRegConst-SP': self.set_tikhonov_reg_const,
            'InvRespMatNormMode-Sel': self.set_invrespmat_normmode,
            'MeasRespMat-Cmd': self.set_respmat_meas_state,
            'MeasRespMatKickCH-SP': _part(self.set_respmat_meas_kick, 'ch'),
            'MeasRespMatKickCV-SP': _part(self.set_respmat_meas_kick, 'cv'),
            'MeasRespMatKickRF-SP': _part(self.set_respmat_meas_kick, 'rf'),
            'MeasRespMatWait-SP': self.set_respmat_meas_wait_time,
        }

        # configuration scanning
        self.quit = False
        self.scanning = False
        self.thread_check_corrs_configs = _epics.ca.CAThread(
            target=self._check_corrs_configs, daemon=True)
        self.thread_check_corrs_configs.start()
        self.thread_check_ctrls_configs = _epics.ca.CAThread(
            target=self._check_ctrls_configs, daemon=True)
        self.thread_check_ctrls_configs.start()

    def init_database(self):
        """Set initial PV values."""
        pvn2vals = {
            'LoopState-Sel': self._loop_state,
            'LoopState-Sts': self._loop_state,
            'LoopGainH-SP': self._loop_gain_h,
            'LoopGainH-RB': self._loop_gain_h,
            'LoopGainH-Mon': self._loop_gain_mon_h,
            'LoopGainV-SP': self._loop_gain_v,
            'LoopGainV-RB': self._loop_gain_v,
            'LoopGainV-Mon': self._loop_gain_mon_v,
            'LoopMaxOrbDistortion-SP': self._loop_max_orb_dist,
            'LoopMaxOrbDistortion-RB': self._loop_max_orb_dist,
            'LoopMaxOrbDistortionEnbl-Sel': self._loop_max_orb_dist_enbl,
            'LoopMaxOrbDistortionEnbl-Sts': self._loop_max_orb_dist_enbl,
            'LoopPacketLossDetecEnbl-Sel': self._loop_packloss_detec_enbl,
            'LoopPacketLossDetecEnbl-Sts': self._loop_packloss_detec_enbl,
            'CorrStatus-Mon': self._corr_status,
            'CorrConfig-Cmd': 0,
            'CorrSetPwrStateOn-Cmd': 0,
            'CorrSetOpModeManual-Cmd': 0,
            'CorrSetAccFreezeDsbl-Cmd': 0,
            'CorrSetAccFreezeEnbl-Cmd': 0,
            'CorrSetAccClear-Cmd': 0,
            'CorrSetCurrZero-Cmd': 0,
            'CorrSetCurrZeroDuration-SP': self._corr_setcurrzero_dur,
            'CorrSetCurrZeroDuration-RB': self._corr_setcurrzero_dur,
            'CHAccSatMax-SP': self._ch_maxacccurr,
            'CHAccSatMax-RB': self._ch_maxacccurr,
            'CVAccSatMax-SP': self._cv_maxacccurr,
            'CVAccSatMax-RB': self._cv_maxacccurr,
            'TimeFrameLen-SP': self._time_frame_len,
            'TimeFrameLen-RB': self._time_frame_len,
            'CtrlrStatus-Mon': self._fofbctrl_status,
            'CtrlrConfBPMId-Cmd': 0,
            'CtrlrSyncNet-Cmd': 0,
            'CtrlrSyncUseEnblList-Sel': self._fofbctrl_syncuseenbllist,
            'CtrlrSyncUseEnblList-Sts': self._fofbctrl_syncuseenbllist,
            'CtrlrSyncEnblList-Mon': self._fofbctrl_syncenbllist,
            'CtrlrSyncRefOrb-Cmd': 0,
            'CtrlrSyncTFrameLen-Cmd': 0,
            'CtrlrConfBPMLogTrg-Cmd': 0,
            'CtrlrSyncMaxOrbDist-Cmd': 0,
            'CtrlrSyncPacketLossDetec-Cmd': 0,
            'CtrlrReset-Cmd': 0,
            'CtrlrDsblSYSIDExc-Cmd': 0,
            'FOFBAccDecimation-Sel': self._corr_accdec_enm,
            'FOFBAccDecimation-Sts': self._corr_accdec_enm,
            'FOFBAccDecimation-SP': self._corr_accdec_val,
            'FOFBAccDecimation-RB': self._corr_accdec_val,
            'MinSingValue-SP': self._min_sing_val,
            'MinSingValue-RB': self._min_sing_val,
            'TikhonovRegConst-SP': self._tikhonov_reg_const,
            'TikhonovRegConst-RB': self._tikhonov_reg_const,
            'InvRespMatNormMode-Sel': self._invrespmat_normmode,
            'InvRespMatNormMode-Sts': self._invrespmat_normmode,
            'MeasRespMat-Cmd': 0,
            'MeasRespMat-Mon': self._const.MeasRespMatMon.Idle,
            'MeasRespMatKickCH-SP': self._meas_respmat_kick['ch'],
            'MeasRespMatKickCH-RB': self._meas_respmat_kick['ch'],
            'MeasRespMatKickCV-SP': self._meas_respmat_kick['cv'],
            'MeasRespMatKickCV-RB': self._meas_respmat_kick['cv'],
            'MeasRespMatKickRF-SP': self._meas_respmat_kick['rf'],
            'MeasRespMatKickRF-RB': self._meas_respmat_kick['rf'],
            'MeasRespMatWait-SP': self._meas_respmat_wait,
            'MeasRespMatWait-RB': self._meas_respmat_wait,
        }
        for pvn, val in pvn2vals.items():
            self.run_callbacks(pvn, val)

        # load autosave data
        # enable lists
        for dev in ['bpmx', 'bpmy', 'ch', 'cv', 'rf']:
            okl = self._load_enbllist(dev)
            pvn = f'{dev.upper()}EnblList-SP' if dev != 'rf' else 'UseRF-Sel'
            enb = self._enable_lists[dev]
            pvv = enb if dev != 'rf' else bool(enb)
            self.run_callbacks(pvn, pvv)
            if not okl:
                self.run_callbacks(
                    pvn.replace('SP', 'RB').replace('Sel', 'Sts'), pvv)
        self._update_fofbctrl_sync_enbllist()
        # matrix
        okm = self._load_respmat()
        self.run_callbacks('RespMat-SP', list(self._respmat.ravel()))
        if not okm:
            self.run_callbacks('RespMat-RB', list(self._respmat.ravel()))
        # ref orbits
        okr = self._load_reforbit()
        for plan in ['x', 'y']:
            pvn = f'RefOrb{plan.upper()}-SP'
            pvv = getattr(self, f'_reforb_{plan}')
            self.run_callbacks(pvn, pvv)
            if not okr:
                self.run_callbacks(pvn.replace('SP', 'RB'), pvv)
        self._update_log('Started.')
        self._init = True

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        value = None
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        _log.info('Write received for: %s --> %s', reason, str(value))
        if reason in self.map_pv2write.keys():
            status = self.map_pv2write[reason](value)
            _log.info('%s Write for: %s --> %s',
                      str(status).upper(), reason, str(value))
            return status
        _log.warning('PV %s does not have a set function.', reason)
        return False

    @property
    def fastcorrs_dev(self):
        """Fast corrector device."""
        return self._corrs_dev

    @property
    def rf_dev(self):
        """RF device."""
        return self._rf_dev

    @property
    def llfofb_dev(self):
        """LL FOFB device."""
        return self._llfofb_dev

    @property
    def havebeam(self):
        """Return if there is stored beam."""
        if self._tests:
            return True
        return self._havebeam_pv.connected and self._havebeam_pv.value

    # --- loop control ---

    def set_loop_state(self, value, abort=False):
        """Set loop state."""
        if not 0 <= value < len(_ETypes.OPEN_CLOSED):
            return False

        self._loop_state_lastsp = value
        if value:
            if not self.havebeam:
                self._update_log('ERR:Do not have stored beam. Aborted.')
                return False
            if _np.any([pvo.value for pvo in self._intlk_pvs]):
                self._update_log('ERR:Reset interlocks before closing')
                self._update_log('ERR:the loop.')
                return False

        if self._thread_loopstate is not None and \
                self._thread_loopstate.is_alive():
            self._update_log('WARN:Interrupting Loop Enable thread...')
            self._abort_thread = True
            self._thread_loopstate.join()

        self._thread_loopstate = _epics.ca.CAThread(
            target=self._thread_set_loop_state,
            args=[value, abort], daemon=True)
        self._thread_loopstate.start()
        return True

    def _thread_set_loop_state(self, value, abort):
        if value:  # closing the loop
            # set gains to zero, recalculate gains and coeffs
            self._update_log('Setting Loop Gain to zero...')
            self._loop_gain_mon_h, self._loop_gain_mon_v = 0, 0
            self._calc_corrs_coeffs(log=False)
            # set and wait corrector gains and coeffs to zero
            if not self._set_corrs_coeffs(log=False):
                return
            self._update_log('Waiting for coefficients and gains...')
            if not self._wait_coeffs_and_gains():
                self.run_callbacks('LoopState-Sel', self._loop_state)
                return

            if self._check_abort_thread():
                return

            # close the loop
            self._update_log('...done. Closing the loop...')
            self._loop_state = value
            if not self._check_set_corrs_opmode():
                self._loop_state = self._const.LoopState.Open
                self.run_callbacks('LoopState-Sel', self._loop_state)
                return
            self.run_callbacks('LoopState-Sts', self._loop_state)

            if self._check_abort_thread():
                return

            # do ramp up
            self._update_log('...done. Starting Loop Gain ramp up...')
            if self._do_loop_gain_ramp(ramp='up'):
                self._update_log('LoopGain ramp up finished!')

        else:  # opening the loop
            # do ramp down
            self._update_log('Starting Loop Gain ramp down...')
            if self._do_loop_gain_ramp(ramp='down', abort=abort):
                self._update_log('Loop Gain ramp down finished!')

            if self._check_abort_thread():
                return

            # open the loop
            self._loop_state = value
            self._check_set_corrs_opmode()
            self.run_callbacks('LoopState-Sts', self._loop_state)

    def _do_loop_gain_ramp(self, ramp='up', abort=False):
        xdata = _np.linspace(0, 1, self._const.LOOPGAIN_RMP_NPTS)
        power = 1
        if ramp == 'up':
            ydata = xdata**power
            ydata_h = ydata * self._loop_gain_h
            ydata_v = ydata * self._loop_gain_v
        else:
            ydata = (1-xdata)**power
            ydata_h = ydata * self._loop_gain_mon_h
            ydata_v = ydata * self._loop_gain_mon_v
        for i in range(self._const.LOOPGAIN_RMP_NPTS):
            if not self.havebeam or abort:
                if not self.havebeam:
                    self._update_log('ERR:There is no beam stored.')
                self._update_log('WARN:Gain ramp aborted.')
                self._update_log('WARN:Setting gain to zero.')
                self._loop_gain_mon_h, self._loop_gain_mon_v = 0, 0
                self.run_callbacks('LoopGainH-Mon', self._loop_gain_mon_h)
                self.run_callbacks('LoopGainV-Mon', self._loop_gain_mon_v)
                self._calc_corrs_coeffs()
                self._set_corrs_coeffs()
                self._wait_coeffs_and_gains()
                return False
            if self._check_abort_thread():
                return False

            self._loop_gain_mon_h = ydata_h[i]
            self._loop_gain_mon_v = ydata_v[i]
            self.run_callbacks('LoopGainH-Mon', self._loop_gain_mon_h)
            self.run_callbacks('LoopGainV-Mon', self._loop_gain_mon_v)
            self._update_log(
                f'{i+1:02d}/{len(ydata):02d} -> Loop Gain: '
                f'H={ydata_h[i]:.3f}, V={ydata_v[i]:.3f}')
            self._calc_corrs_coeffs(log=False)
            self._set_corrs_coeffs(log=False)
            _t0 = _time.time()
            if not self._wait_coeffs_and_gains():
                return False
            _td = 1/self._const.LOOPGAIN_RMP_FREQ - (_time.time() - _t0)
            if _td > 0:
                _time.sleep(_td)
        return True

    def _wait_coeffs_and_gains(self):
        _t0 = _time.time()
        while _time.time() - _t0 < self._const.DEF_TIMEOUT:
            _time.sleep(self._const.DEF_TIMESLEEP)
            if self._corrs_dev.check_invrespmat_row(self._pscoeffs) and \
                    self._corrs_dev.check_fofbacc_gain(self._psgains):
                return True
            if self._check_abort_thread():
                return False
        self._update_log('ERR:Timed out waiting for correctors to')
        self._update_log('ERR:implement gains and coefficients.')
        return False

    def _check_abort_thread(self):
        if self._abort_thread:
            self._update_log('WARN:Loop state thread aborted.')
            self._abort_thread = False
            return True
        return False

    def set_loop_gain(self, plane, value):
        """Set loop gain."""
        if not -2**3 <= value <= 2**3-1:
            return False

        if self._thread_loopstate is not None and \
                self._thread_loopstate.is_alive():
            self._update_log('ERR:Wait for Loop Gain ramp before ')
            self._update_log('ERR:setting new value.')
            return False

        setattr(self, '_loop_gain_' + plane, value)

        # if loop closed, calculate new gains and coefficients
        if self._loop_state:
            setattr(self, '_loop_gain_mon_' + plane, value)
            self.run_callbacks(f'LoopGain{plane.upper()}-Mon', value)
            self._calc_corrs_coeffs()
            self._set_corrs_coeffs()

        self._update_log(f'Changed Loop Gain {plane.upper()} to {value}.')
        self.run_callbacks(f'LoopGain{plane.upper()}-RB', value)
        return True

    def set_loop_max_orbit_dist(self, value):
        """Set maximum orbit distortion threshold."""
        if not 0 <= value <= 10000:
            return False
        if not self._check_fofbctrl_connection():
            return False

        self._loop_max_orb_dist = value
        self._update_log('Setting orbit distortion threshold...')
        if self._llfofb_dev.set_max_orb_distortion(
                value=value * self._const.CONV_UM_2_NM,
                timeout=self._const.DEF_TIMEWAIT):
            self._update_log('...done!')
        else:
            self._update_log('ERR:Failed to set threshold.')

        self.run_callbacks('LoopMaxOrbDistortion-RB', value)
        return True

    def set_loop_max_orbit_dist_enbl(self, value):
        """Set orbit distortion detection enable status."""
        if not self._check_fofbctrl_connection():
            return False

        act = ('En' if value else 'Dis')
        self._loop_max_orb_dist_enbl = value
        self._update_log(act+'abling orbit distortion detection...')
        if self._llfofb_dev.set_max_orb_distortion_enbl(
                value=value, timeout=self._const.DEF_TIMEWAIT):
            self._update_log('...done!')
        else:
            self._update_log('ERR:Failed to '+act+'able detection.')

        self.run_callbacks('LoopMaxOrbDistortionEnbl-Sts', value)
        return True

    def set_loop_packloss_detec_enbl(self, value):
        """Set packet loss detection enable status."""
        if not self._check_fofbctrl_connection():
            return False

        act = ('En' if value else 'Dis')
        self._loop_packloss_detec_enbl = value
        self._update_log(act+'abling packet loss detection...')
        if self._llfofb_dev.set_min_bpm_count_enbl(
                value=value, timeout=self._const.DEF_TIMEWAIT):
            self._update_log('...done!')
        else:
            self._update_log('ERR:Failed to '+act+'able detection.')

        self.run_callbacks('LoopPacketLossDetecEnbl-Sts', value)
        return True

    # --- devices configuration ---

    def cmd_corr_configure(self, _):
        """Configure corrector command."""
        self._update_log('Received configure corrector command...')
        if not self._check_corr_connection():
            return False

        # saturation limits
        self._update_log('Setting corrector saturation limits...')
        chn, chl = self._const.ch_names, self._ch_maxacccurr
        cvn, cvl = self._const.cv_names, self._cv_maxacccurr
        self._corrs_dev.set_fofbacc_satmax(chl, psnames=chn)
        self._corrs_dev.set_fofbacc_satmin(-chl, psnames=chn)
        self._corrs_dev.set_fofbacc_satmax(cvl, psnames=cvn)
        self._corrs_dev.set_fofbacc_satmin(-cvl, psnames=cvn)
        self._update_log('...done!')
        # pwrstate
        self._check_set_corrs_pwrstate()
        # opmode
        self._check_set_corrs_opmode()
        # fofbacc_freeze
        self._set_corrs_fofbacc_freeze()
        # matrix coefficients
        self._set_corrs_coeffs()

        self._update_log('Correctors configuration done!')
        return True

    def cmd_corr_pwrstate_on(self, _):
        """Set all corrector pwrstate to on."""
        self._update_log('Received set corrector pwrstate to on...')
        if not self._check_corr_connection():
            return False

        self._update_log('Setting all corrector pwrstate to on...')
        self._corrs_dev.set_pwrstate(self._const.OffOn.On)
        self._update_log('...done!')
        return True

    def cmd_corr_pwrstate_off(self, _):
        """Set all corrector pwrstate to off."""
        self._update_log('Received set corrector pwrstate to off...')
        if not self._check_corr_connection():
            return False

        self._update_log('Setting all corrector pwrstate to off...')
        self._corrs_dev.set_pwrstate(self._const.OffOn.Off)
        self._update_log('...done!')
        return True

    def cmd_corr_opmode_manual(self, _):
        """Set all corrector opmode to manual."""
        self._update_log('Received set corrector opmode to manual...')
        if not self._check_corr_connection():
            return False

        self._update_log('Setting all corrector opmode to manual...')
        self._corrs_dev.set_opmode(self._corrs_dev.OPMODE_STS.manual)
        self._update_log('...done!')
        return True

    def cmd_corr_accfreeze_dsbl(self, _):
        """Set all corrector accumulator freeze state to Dsbl."""
        self._update_log('Received set corrector AccFreeze to Dsbl...')
        if not self._check_corr_connection():
            return False

        self._update_log('Setting AccFreeze to Dsbl...')
        self._corrs_dev.set_fofbacc_freeze(self._const.DsblEnbl.Dsbl)
        self._update_log('...done!')
        return True

    def cmd_corr_accfreeze_enbl(self, _):
        """Set all corrector accumulator freeze state to Enbl."""
        self._update_log('Received set corrector AccFreeze to Enbl...')
        if not self._check_corr_connection():
            return False

        self._update_log('Setting AccFreeze to Enbl...')
        self._corrs_dev.set_fofbacc_freeze(self._const.DsblEnbl.Enbl)
        self._update_log('...done!')
        return True

    def cmd_corr_accclear(self, _):
        """Clear all corrector accumulator."""
        self._update_log('Received clear all corrector accumulator...')
        if not self._check_corr_connection():
            return False

        self._update_log('Sending clear accumulator command...')
        self._corrs_dev.cmd_fofbacc_clear()
        self._update_log('...done!')
        return True

    def cmd_corr_currzero(self, _):
        """Set all corrector current to zero."""
        self._update_log('Received set corrector current to zero...')
        if not self._check_corr_connection():
            return False
        if self._thread_currzero is not None and \
                self._thread_currzero.is_alive():
            self._update_log('ERR:Current ramp down already in progress.')
            return False

        self._thread_currzero = _epics.ca.CAThread(
            target=self._thread_corr_currzero, daemon=True)
        self._thread_currzero.start()
        return True

    def set_corr_currzero_duration(self, value):
        """Set corrector ramp down current duration."""
        if not 0 <= value <= 1000:
            return False

        self._corr_setcurrzero_dur = value
        self._update_log('Changed corrector current ramp duration.')
        if self._thread_currzero is not None and \
                self._thread_currzero.is_alive():
            self._abort_thread_currzero = True
            self._thread_currzero.join()
            self._thread_currzero = _epics.ca.CAThread(
                target=self._thread_corr_currzero, daemon=True)
            self._thread_currzero.start()
        self.run_callbacks('CorrSetCurrZeroDuration-RB', value)
        return True

    def set_corr_accdec(self, option, value):
        """Set corrector accumulator decimation."""
        if self._corr_accdec_enm != self._const.DecOpt.Custom and \
                option == 'value':
            return False

        if option == 'enum':
            if value == self._const.DecOpt.FOFB:
                dec = 1
            elif value == self._const.DecOpt.Monit:
                if self._auxbpm.connected:
                    monit = self._auxbpm['INFOMONITRate-RB']
                    fofb = self._auxbpm['INFOFOFBRate-RB']
                    dec = monit // fofb
                else:
                    self._update_log('WARN:Could not read decimation from BPM')
                    self._update_log('WARN:rates. Using value 4600.')
                    dec = 4600
            else:
                dec = self._corr_accdec_val
            self._corr_accdec_enm = value
            self.run_callbacks('FOFBAccDecimation-Sts', value)
            self.run_callbacks('FOFBAccDecimation-SP', dec)
        else:
            dec = value
        self._corr_accdec_val = dec
        self.run_callbacks('FOFBAccDecimation-RB', dec)

        self._update_log('Setting FOFB Acc decimation...')
        self._corrs_dev.set_fofbacc_decimation(dec)
        self._update_log('...done!')

        return True

    def _thread_corr_currzero(self):
        if self._corrs_dev.check_current(0):
            self._update_log('Current of all correctors already zeroed.')
            return

        self._update_log('Sending all corrector current to zero...')

        init_curr = self._corrs_dev.current
        npts = int(self._corr_setcurrzero_dur*self._const.CURRZERO_RMP_FREQ)
        if npts != 0:
            xdata = _np.linspace(1, 0, npts)
            for idx, step in enumerate(xdata):
                curr = init_curr * step
                if self._check_thread_currzero_abort():
                    self._update_log('...aborted.')
                    return
                self._update_log(
                    f'{idx+1:02d}/{len(xdata):02d} -> Current={100*step:.1f}%')
                self._corrs_dev.set_current(curr)
                _time.sleep(1/self._const.CURRZERO_RMP_FREQ)

        self._corrs_dev.set_current(0)
        self._update_log('...done!')

    def _check_thread_currzero_abort(self):
        if self._abort_thread_currzero:
            self._abort_thread_currzero = False
            return True
        return False

    def set_corr_accsatmax(self, device, value):
        """Set device FOFB accumulator saturation limits."""
        if not self._check_corr_connection():
            return False
        if not 0 <= value <= 0.95:
            return False

        setattr(self, '_'+device+'_maxacccurr', value)
        self._update_log('Setting '+device.upper()+' saturation limits...')
        psnames = getattr(self._const, device+'_names')
        self._corrs_dev.set_fofbacc_satmax(value, psnames=psnames)
        self._corrs_dev.set_fofbacc_satmin(-value, psnames=psnames)
        self._update_log('...done!')

        self._update_log(
            'Changed '+device.upper()+' saturation limits to '+str(value)+'A.')
        self.run_callbacks(device.upper()+'AccSatMax-RB', value)
        return True

    def set_timeframelen(self, value):
        """Set FOFB controllers TimeFrameLen."""
        if not 500 <= value <= 10000:
            return False
        if not self._check_fofbctrl_connection():
            return False

        self._time_frame_len = value
        self._update_log(f'Setting TimeFrameLen to {value}...')
        self._llfofb_dev.set_time_frame_len(
            value=self._time_frame_len, timeout=self._const.DEF_TIMEWAIT)
        self._update_log('...done!')

        self.run_callbacks('TimeFrameLen-RB', self._time_frame_len)
        return True

    def cmd_fofbctrl_confbpmid(self, _):
        """Configure FOFB DCC BPMId command."""
        self._update_log('Received configure FOFB DCC BPMId command...')
        if not self._check_fofbctrl_connection():
            return False
        self._update_log('Checking...')
        if not self._llfofb_dev.bpm_id_configured:
            self._update_log('Configuring DCC BPMIds...')
            if self._llfofb_dev.cmd_config_bpm_id():
                self._update_log('Sent configuration to DCCs.')
            else:
                self._update_log('ERR:Failed to configure DCCs.')
        else:
            self._update_log('FOFB DCC BPMIds already configured.')
        return True

    def cmd_fofbctrl_syncnet(self, _):
        """Sync FOFB net command."""
        self._update_log('Received sync FOFB net command...')
        if not self._check_fofbctrl_connection():
            return False
        self._update_log('Checking...')

        if self._thread_syncnet is not None and \
                self._thread_syncnet.is_alive():
            self._update_log('ERR:Net sync already in progress.')
            return False

        self._thread_syncnet = _epics.ca.CAThread(
            target=self._thread_fofbctrl_syncnet, daemon=True)
        self._thread_syncnet.start()
        return True

    def _thread_fofbctrl_syncnet(self):
        steps = [
            self._dsbl_fofbctrl_minbpmcnt_enbl,
            self._do_fofbctrl_syncnet,
            self._wait_fofbctrl_netsync,
            self._conf_fofbctrl_minbpmcnt_enbl,
        ]
        for func in steps:
            if not func():
                break

    def cmd_fofbctrl_syncreforb(self, _):
        """Sync FOFB RefOrb command."""
        self._update_log('Received sync FOFB RefOrb command...')
        if not self._check_fofbctrl_connection():
            return False
        self._update_log('Checking...')
        if not self._llfofb_dev.check_reforbx(self._reforbhw_x) or not \
                self._llfofb_dev.check_reforby(self._reforbhw_y):
            self._update_log('Syncing FOFB RefOrb...')
            self._llfofb_dev.set_reforbx(self._reforbhw_x)
            self._llfofb_dev.set_reforby(self._reforbhw_y)
            self._update_log('...done!')
        else:
            self._update_log('FOFB RefOrb already synced.')
        return True

    def cmd_fofbctrl_synctframelen(self, _):
        """Sync FOFB controllers TimeFrameLen command."""
        self._update_log('Received configure FOFB controllers')
        self._update_log('TimeFrameLen command... Checking...')
        if not self._check_fofbctrl_connection():
            return False
        timeframe = self._time_frame_len
        if not _np.all(self._llfofb_dev.time_frame_len == timeframe):
            self._update_log('Configuring TimeFrameLen PVs...')
            if self._llfofb_dev.set_time_frame_len(timeframe):
                self._update_log('...done!')
            else:
                self._update_log('ERR:Failed to configure TimeFrameLen.')
        else:
            self._update_log('TimeFrameLen PVs already configured.')
        return True

    def cmd_fofbctrl_confbpmlogtrg(self, _):
        """Configure BPM logical triggers command."""
        self._update_log('Received configure BPM Logical')
        if not self._check_fofbctrl_connection():
            return False
        self._update_log('triggers command... Checking...')
        if not self._llfofb_dev.bpm_trigs_configured:
            self._update_log('Configuring BPM logical triggers...')
            if self._llfofb_dev.cmd_config_bpm_trigs():
                self._update_log('...done!')
            else:
                self._update_log('ERR:Failed to configure BPM log.trigs.')
        else:
            self._update_log('BPM logical triggers already configured.')
        return True

    def set_fofbctrl_syncuseenablelist(self, value):
        """Set whether to use or not BPMEnblList in sync."""
        if not self._check_fofbctrl_connection():
            return False
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False

        self._fofbctrl_syncuseenbllist = value
        self._update_fofbctrl_sync_enbllist()
        self._conf_fofbctrl_packetlossdetec()

        self._update_log('Changed sync net command to ')
        self._update_log(('' if value else 'not ')+'use BPM EnableList.')
        self.run_callbacks('CtrlrSyncUseEnblList-Sts', value)
        return True

    def cmd_fofbctrl_syncmaxorbdist(self, _):
        """Sync FOFB controllers orbit distortion detection command."""
        self._update_log('Received sync FOFB controllers orbit')
        self._update_log('distortion detection command...Checking...')
        if not self._check_fofbctrl_connection():
            return False

        tout = self._const.DEF_TIMEWAIT

        # threshold
        thres = self._loop_max_orb_dist * self._const.CONV_UM_2_NM
        if not _np.all(self._llfofb_dev.max_orb_distortion == thres):
            self._update_log('Setting MaxOrbDistortion PVs...')
            if self._llfofb_dev.set_max_orb_distortion(thres, timeout=tout):
                self._update_log('...done!')
            else:
                self._update_log('ERR:Failed to sync threshold.')
        else:
            self._update_log('MaxOrbDistortion PVs already configured.')

        # enable status
        sts = self._loop_max_orb_dist_enbl
        if not _np.all(self._llfofb_dev.max_orb_distortion_enbl == sts):
            self._update_log('Setting MaxOrbDistortionEnbl PVs...')
            if self._llfofb_dev.set_max_orb_distortion_enbl(sts, timeout=tout):
                self._update_log('...done!')
            else:
                self._update_log('ERR:Failed to sync enable status.')
        else:
            self._update_log('MaxOrbDistortionEnbl PVs already configured.')
        return True

    def cmd_fofbctrl_syncpacklossdet(self, _):
        """Sync FOFB controllers packet loss detection command."""
        self._update_log('Received sync FOFB controllers packet')
        self._update_log('loss detection command...Checking...')
        if not self._check_fofbctrl_connection():
            return False
        self._conf_fofbctrl_packetlossdetec()
        return True

    def cmd_fofbctrl_reset(self, _):
        """Reset FOFB controllers interlock command."""
        self._update_log('Received reset FOFB controllers command...')
        if not self._check_fofbctrl_connection():
            return False
        self._do_fofbctrl_reset()
        return True

    def cmd_fofbctrl_dsblsysid(self, _):
        """Disable FOFB controllers system identification excitation."""
        self._update_log('Received disable FOFB controllers SYSID...')
        if not self._check_fofbctrl_connection():
            return False
        self._update_log('Disabling SYSID excitation...')
        self._llfofb_dev.cmd_dsbl_sysid_exc()
        self._update_log('...done!')
        return True

    # --- reference orbit ---

    def set_reforbit(self, plane, value):
        """Set reference orbit."""
        self._update_log(f'Setting New RefOrb{plane.upper()}...')

        # check size
        ref = _np.array(value, dtype=float)
        if ref.size != self._const.reforb_size:
            self._update_log('ERR: Wrong RefOrb Size.')
            return False

        # set internal states and LLFOFB reforb
        # physical units
        setattr(self, '_reforb_' + plane.lower(), ref)
        # hardware units
        refhw = ref * self._const.CONV_UM_2_NM
        refhw = _np.round(refhw)  # round, low level expect it to be int
        refhw = _np.roll(refhw, 1)  # make BPM 01M1 the first element
        setattr(self, '_reforbhw_' + plane.lower(), refhw)

        # do not set reforb to controllers and save to file in initialization
        if self._init:
            # set reforb to FOFB controllers
            func = getattr(self._llfofb_dev, 'set_reforb' + plane.lower())
            func(refhw)

            # save to autosave file
            self._save_reforbit()

        # update readback PV
        self.run_callbacks(f'RefOrb{plane.upper()}-RB', list(ref.ravel()))
        self.run_callbacks(f'RefOrbHw{plane.upper()}-Mon', refhw)
        self._update_log('...done!')
        return True

    def _load_reforbit(self):
        filename = self._const.reforb_fname
        if not _os.path.isfile(filename):
            return False
        self._update_log('Loading RefOrbits from file...')
        refx, refy = _np.loadtxt(filename, unpack=True)
        okx = self.set_reforbit('x', refx)
        oky = self.set_reforbit('y', refy)
        if all([okx, oky]):
            msg = 'Loaded RefOrbits!'
        else:
            msg = 'ERR:Problem loading RefOrbits from file.'
        self._update_log(msg)
        return 'ERR' not in msg

    def _save_reforbit(self):
        filename = self._const.reforb_fname
        refx, refy = self._reforb_x, self._reforb_y
        orbs = _np.array([refx, refy]).T
        try:
            path = _os.path.split(filename)[0]
            _os.makedirs(path, exist_ok=True)
            _np.savetxt(filename, orbs)
        except FileNotFoundError:
            self._update_log('WARN:Could not save ref.orbit to file.')

    # --- matrix manipulation ---

    def set_respmat(self, value):
        """Set response matrix."""
        self._update_log('Setting New RespMat...')

        # check size
        mat = _np.array(value, dtype=float)
        matb = mat
        if mat.size != self._const.matrix_size:
            self._update_log(
                f'ERR: Wrong RespMat Size ({mat.size}, '
                f'expected {self._const.matrix_size}).')
            return False

        # reshape
        nrc = self._const.nr_corrs
        mat = _np.reshape(mat, [-1, nrc])

        # check if matrix is invertible
        old_ = self._respmat.copy()
        self._respmat = mat
        if not self._calc_matrices():
            self._respmat = old_
            return False

        # if ok, save matrix
        if self._init:
            self._save_respmat(matb)

        # update readback pv
        self.run_callbacks('RespMat-RB', list(self._respmat.ravel()))

        return True

    def _load_respmat(self):
        filename = self._const.respmat_fname
        if not _os.path.isfile(filename):
            return False
        self._update_log('Loading RespMat from file...')
        if self.set_respmat(_np.loadtxt(filename)):
            msg = 'Loaded RespMat!'
        else:
            msg = 'ERR:Problem loading RespMat from file.'
        self._update_log(msg)
        return 'ERR' not in msg

    def _save_respmat(self, mat):
        try:
            filename = self._const.respmat_fname
            path = _os.path.split(filename)[0]
            _os.makedirs(path, exist_ok=True)
            _np.savetxt(filename, mat)
        except FileNotFoundError:
            self._update_log('WARN:Could not save matrix to file.')

    def set_enbllist(self, device, value):
        """Set enable list for device."""
        if self._loop_state:
            self._update_log('ERR:Open loop before continuing.')
            return False
        if self._thread_enbllist is not None and \
                self._thread_enbllist.is_alive():
            self._abort_thread_enbllist = True
            self._thread_enbllist.join()
        self._update_log('Setting {0:s} EnblList'.format(device.upper()))

        # check size
        bkup = self._enable_lists[device]
        new = _np.array(value, dtype=bool)
        if bkup.size != new.size:
            self._update_log(
                'ERR: Wrong {0:s} EnblList size.'.format(device.upper()))
            return False

        self._enable_lists[device] = new

        # do not set enable lists and save to file in initialization
        if self._init:
            # check if matrix is invertible
            if not self._calc_matrices():
                self._enable_lists[device] = bkup
                return False

            # handle devices enable configuration
            self._thread_enbllist = _epics.ca.CAThread(
                target=self._handle_devices_enblconfig, args=[device, ],
                daemon=True)
            self._thread_enbllist.start()

            # save to autosave files
            self._save_enbllist(device, _np.array([value], dtype=bool))

        # update readback pv
        if device == 'rf':
            self.run_callbacks('UseRF-Sts', bool(value))
        else:
            self.run_callbacks(device.upper()+'EnblList-RB', new)

        return True

    def _handle_devices_enblconfig(self, device):
        if device in ['ch', 'cv']:
            if self._check_corr_connection():
                self._check_set_corrs_opmode()
        elif device in ['bpmx', 'bpmy']:
            self._update_fofbctrl_sync_enbllist()
            if self._check_fofbctrl_connection():
                steps = [
                    self._dsbl_fofbctrl_minbpmcnt_enbl,
                    self._do_fofbctrl_syncnet,
                    self._wait_fofbctrl_netsync,
                    self._conf_fofbctrl_minbpmcnt,
                    self._conf_fofbctrl_minbpmcnt_enbl,
                ]
                for func in steps:
                    if self._check_thread_enbllist_abort():
                        break
                    if not func():
                        break

    def _check_thread_enbllist_abort(self):
        if self._abort_thread_enbllist:
            self._abort_thread_enbllist = False
            return True
        return False

    @property
    def bpm_enbllist(self):
        """BPM enable list."""
        enbl = self._enable_lists
        return _np.hstack([enbl['bpmx'], enbl['bpmy']])

    @property
    def corr_enbllist(self):
        """Corrector enable list."""
        enbl = self._enable_lists
        return _np.hstack([enbl['ch'], enbl['cv'], enbl['rf']])

    def _load_enbllist(self, device):
        filename = getattr(self._const, device+'enbl_fname')
        if not _os.path.isfile(filename):
            return
        okl = self.set_enbllist(device, _np.loadtxt(filename))
        if okl:
            msg = f'Loaded {device} enable list!'
        else:
            msg = f'ERR:Problem loading {device} enable list from file.'
        self._update_log(msg)
        return okl

    def _save_enbllist(self, device, value):
        try:
            filename = getattr(self._const, device+'enbl_fname')
            path = _os.path.split(filename)[0]
            _os.makedirs(path, exist_ok=True)
            _np.savetxt(filename, value)
        except FileNotFoundError:
            self._update_log(
                f'WARN:Could not save {device} enable list to file.')

    def set_min_sing_value(self, value):
        """Set minimum singular values."""
        bkup = self._min_sing_val
        self._min_sing_val = float(value)
        if not self._calc_matrices():
            self._min_sing_val = bkup
            return False
        self.run_callbacks('MinSingValue-RB', self._min_sing_val)
        return True

    def set_tikhonov_reg_const(self, value):
        """Set Tikhonov regularization constant."""
        bkup = self._tikhonov_reg_const
        self._tikhonov_reg_const = float(value)
        if not self._calc_matrices():
            self._tikhonov_reg_const = bkup
            return False
        self.run_callbacks('TikhonovRegConst-RB', self._tikhonov_reg_const)
        return True

    def set_invrespmat_normmode(self, value):
        """Set corrector coefficients normalization mode."""
        if not 0 <= value < len(_ETypes.GLOB_INDIV):
            return False

        self._update_log(
            'Changed normalization mode to ' + str(
                self._const.GlobIndiv._fields[value])+'.')
        self._invrespmat_normmode = value

        self._calc_corrs_coeffs()
        self._set_corrs_coeffs()

        self.run_callbacks(
            'InvRespMatNormMode-Sts', self._invrespmat_normmode)
        return True

    def _calc_matrices(self):
        self._update_log('Calculating Inverse Matrix...')

        if not self._check_corr_connection():
            return False

        selbpm = self.bpm_enbllist
        if not any(selbpm):
            self._update_log('ERR: No BPM selected in EnblList')
            return False
        selcorr = self.corr_enbllist
        if not any(selcorr):
            self._update_log('ERR: No Corrector selected in EnblList')
            return False

        # apply device selection
        selmat = selbpm[:, None] * selcorr[None, :]
        if selmat.size != self._respmat.size:
            return False
        mat = self._respmat.copy()
        mat = mat[selmat]
        mat = _np.reshape(mat, [sum(selbpm), sum(selcorr)])

        # calculate SVD
        try:
            _uo, _so, _vo = _np.linalg.svd(mat, full_matrices=False)
        except _np.linalg.LinAlgError():
            self._update_log('ERR: Could not calculate SVD')
            return False

        # handle singular values
        # select singular values greater than minimum
        idcs = _so > self._min_sing_val
        _sr = _so[idcs]
        nrs = _np.sum(idcs)
        if not nrs:
            self._update_log('ERR: All Singular Values below minimum.')
            return False
        # apply Tikhonov regularization
        regc = self._tikhonov_reg_const
        regc *= regc
        inv_s = _np.zeros(_so.size, dtype=float)
        inv_s[idcs] = _sr/(_sr*_sr + regc)
        # calculate processed singular values
        _sp = _np.zeros(_so.size, dtype=float)
        _sp[idcs] = 1/inv_s[idcs]

        # check if inverse matrix is valid
        invmat = _np.dot(_vo.T*inv_s, _uo.T)
        if _np.any(_np.isnan(invmat)) or _np.any(_np.isinf(invmat)):
            self._update_log('ERR: Inverse contains nan or inf.')
            return False

        # reconstruct filtered and regularized matrix in physical units
        matr = _np.dot(_uo*_sp, _vo)

        # convert matrix to hardware units
        str2curr = _np.r_[self._corrs_dev.strength_2_current_factor, 1.0]
        currgain = _np.r_[self._corrs_dev.curr_gain, 1.0]
        if _np.any(str2curr == 0) or _np.any(currgain == 0):
            self._update_log('ERR:Could not calculate hardware unit')
            self._update_log('ERR:matrix, CurrGain or "urad to A" ')
            self._update_log('ERR:factor have zero values.')
            return False
        # unit convertion: um/urad (1)-> nm/urad (2)-> nm/A (3)-> nm/counts
        matc = matr * self._const.CONV_UM_2_NM
        matc = matc / str2curr[selcorr]
        matc = matc * currgain[selcorr]

        # obtain pseudoinverse
        # calculate SVD for converted matrix
        _uc, _sc, _vc = _np.linalg.svd(matc, full_matrices=False)
        # handle singular value selection
        idcsc = _sc/_sc.max() >= self._const.SINGVALHW_THRS
        inv_sc = _np.zeros(_so.size, dtype=float)
        inv_sc[idcsc] = 1/_sc[idcsc]
        # calculate pseudoinverse of converted matrix from SVD
        invmatc = _np.dot(_vc.T*inv_sc, _uc.T)

        # update internal states and PVs
        sing_vals = _np.zeros(self._const.nr_svals, dtype=float)
        sing_vals[:_so.size] = _so
        self.run_callbacks('SingValuesRaw-Mon', sing_vals)

        sing_vals = _np.zeros(self._const.nr_svals, dtype=float)
        sing_vals[:_sp.size] = _sp
        self.run_callbacks('SingValues-Mon', sing_vals)
        self.run_callbacks('NrSingValues-Mon', nrs)

        self._invrespmat = _np.zeros(self._respmat.shape, dtype=float).T
        self._invrespmat[selmat.T] = invmat.ravel()
        self.run_callbacks('InvRespMat-Mon', list(self._invrespmat.ravel()))

        respmat_proc = _np.zeros(self._respmat.shape, dtype=float)
        respmat_proc[selmat] = matr.ravel()
        self.run_callbacks('RespMat-Mon', list(respmat_proc.ravel()))

        sing_vals = _np.zeros(self._const.nr_svals, dtype=float)
        sing_vals[:_sc.size] = _sc
        self.run_callbacks('SingValuesHw-Mon', sing_vals)

        self._invrespmatconv = _np.zeros(self._respmat.shape, dtype=float).T
        self._invrespmatconv[selmat.T] = invmatc.ravel()
        self.run_callbacks(
            'InvRespMatHw-Mon', list(self._invrespmatconv.ravel()))

        respmatconv_proc = _np.zeros(self._respmat.shape, dtype=float)
        respmatconv_proc[selmat] = matc.ravel()
        self.run_callbacks('RespMatHw-Mon', list(respmatconv_proc.ravel()))

        # send new matrix to low level FOFB
        self._calc_corrs_coeffs()
        if self._init:
            self._set_corrs_coeffs()

        self._update_log('Ok!')
        return True

    # --- matrix measurement ---

    def set_respmat_meas_state(self, value):
        """Set response matrix measure state."""
        if value == self._const.MeasRespMatCmd.Start:
            return self._start_meas_respmat()
        if value == self._const.MeasRespMatCmd.Stop:
            return self._stop_meas_respmat()
        if value == self._const.MeasRespMatCmd.Reset:
            return self._reset_meas_respmat()
        return True

    def set_respmat_meas_kick(self, plane, value):
        """Set response matrix measure kick value for plane [urad]."""
        self._meas_respmat_kick[plane] = value
        self.run_callbacks('MeasRespMatKick'+plane.upper()+'-RB', value)
        return True

    def set_respmat_meas_wait_time(self, value):
        """Set response matrix measure wait time [s]."""
        self._meas_respmat_wait = value
        self.run_callbacks('MeasRespMatWait-RB', value)
        return True

    def _start_meas_respmat(self):
        if self._loop_state == self._const.LoopState.Closed:
            self._update_log('ERR: Open FOFB loop before continue.')
            return False
        if not self._sofb_check_config():
            self._update_log('ERR: Aborted.')
            return False
        if self._measuring_respmat:
            self._update_log('ERR: Measurement already in progress...')
            return False
        if not self.havebeam:
            self._update_log('ERR: Do not have stored beam. Aborted.')
            return False
        self._update_log('Starting RespMat measurement.')
        self._measuring_respmat = True
        self._meas_respmat_thread = _epics.ca.CAThread(
            target=self._do_meas_respmat, daemon=True)
        self._meas_respmat_thread.start()
        return True

    def _stop_meas_respmat(self):
        if not self._measuring_respmat:
            self._update_log('ERR: No Measurement occurring.')
            return False
        self._update_log('Aborting measurement. Wait...')
        self._measuring_respmat = False
        return True

    def _reset_meas_respmat(self):
        if self._measuring_respmat:
            self._update_log('WARN: measurement in progress...')
            return False
        self._update_log('Reseting measurement status.')
        self.run_callbacks('MeasRespMat-Mon', self._const.MeasRespMatMon.Idle)
        return True

    def _do_meas_respmat(self):
        self.run_callbacks(
            'MeasRespMat-Mon', self._const.MeasRespMatMon.Measuring)
        mat = list()
        enbllist = self.corr_enbllist
        sum_enbld = sum(enbllist)

        orbzero = _np.zeros(len(self.bpm_enbllist), dtype=float)
        for i in range(self._const.nr_corrs):
            if not self._measuring_respmat:
                self.run_callbacks(
                    'MeasRespMat-Mon', self._const.MeasRespMatMon.Aborted)
                self._update_log('Measurement stopped.')
                for _ in range(i, self._const.nr_corrs):
                    mat.append(orbzero)
                break
            if not self.havebeam:
                self.run_callbacks(
                    'MeasRespMat-Mon', self._const.MeasRespMatMon.Aborted)
                self._update_log(
                    'ERR: Cannot Measure, We do not have stored beam!')
                for _ in range(i, self._const.nr_corrs):
                    mat.append(orbzero)
                break
            if not enbllist[i]:
                mat.append(orbzero)
                continue

            if i < self._const.nr_ch + self._const.nr_cv:
                dev = self._corrs_dev[i]
                conv = self._corrs_dev.psconvs[i]
                self._update_log('{0:d}/{1:d} -> {2:s}'.format(
                    i+1, sum_enbld, dev.devname))

                corrtype = 'ch' if 'FCH' in dev.devname else 'cv'
                delta = self._meas_respmat_kick[corrtype]

                orig_kick = dev.strength
                orig_curr = dev.current

                kickp = orig_kick + delta/2
                dev.current = conv.conv_strength_2_current(kickp)
                _time.sleep(self._meas_respmat_wait)
                orbp = self._sofb_get_orbit()

                kickn = orig_kick - delta/2
                dev.current = conv.conv_strength_2_current(kickn)
                _time.sleep(self._meas_respmat_wait)
                orbn = self._sofb_get_orbit()

                dev.current = orig_curr
            elif i < self._const.nr_corrs:
                dev = self.rf_dev
                self._update_log('{0:d}/{1:d} -> {2:s}'.format(
                    i+1, sum_enbld, dev.devname))

                delta = self._meas_respmat_kick['rf']

                orig_freq = dev.frequency

                dev.frequency = orig_freq + delta/2
                _time.sleep(self._meas_respmat_wait)
                orbp = self._sofb_get_orbit()

                dev.frequency = orig_freq - delta/2
                _time.sleep(self._meas_respmat_wait)
                orbn = self._sofb_get_orbit()

                dev.frequency = orig_freq
            mat.append((orbp - orbn)/delta)

        mat = _np.array(mat).T
        self.set_respmat(list(mat.ravel()))
        self.run_callbacks(
            'MeasRespMat-Mon', self._const.MeasRespMatMon.Completed)
        self._measuring_respmat = False
        self._update_log('RespMat Measurement Completed!')

    def _sofb_check_config(self):
        if not self._sisofb_dev.autocorrsts == _Const.LoopState.Open:
            self._update_log('ERR: Open SOFB loop before continue.')
            return False
        if not self._sisofb_dev.opmode_str == 'SlowOrb':
            self._update_log('ERR: SOFBMode is different from SlowOrb.')
            return False
        if not self._sisofb_dev.wait_orb_status_ok(timeout=0.5):
            self._update_log('ERR: SOFB orbit status is not ok.')
            return False
        return True

    def _sofb_get_orbit(self):
        self._sisofb_dev.cmd_reset()
        _time.sleep(self._const.DEF_TIMESLEEP)
        self._sisofb_dev.wait_buffer()
        orbx, orby = self._sisofb_dev.orbx, self._sisofb_dev.orby
        return _np.hstack([orbx, orby])

    # --- callbacks ---

    def _callback_havebeam(self, value, **kws):
        if not value and self._loop_state == self._const.LoopState.Closed:
            self._update_log('FATAL:We do not have stored beam!')
            self._update_log('FATAL:Opening FOFB loop...')
            self.set_loop_state(self._const.LoopState.Open, abort=True)

    def _callback_loopintlk(self, pvname, value, **kws):
        sub = _PVName(pvname).sub[:2]
        old = self._intlk_values[pvname]
        orbdis = _get_bit(value, 0) and not _get_bit(old, 0)
        paclos = _get_bit(value, 1) and not _get_bit(old, 1)
        self._intlk_values[pvname] = value
        if value != 0:
            pref = ('FATAL' if self._loop_state else 'WARN') + \
                ':Ctrlr.' + sub + ' detected '
            if orbdis:
                self._update_log(pref + 'large orb.dist.!')
            if paclos:
                self._update_log(pref + 'packet loss!')

            if self._loop_state != self._const.LoopState.Closed:
                return

            if self._thread_loopstate is None or \
                    (self._thread_loopstate is not None and
                     not self._thread_loopstate.is_alive()) or \
                    (self._thread_loopstate is not None and
                     self._thread_loopstate.is_alive() and
                     self._loop_state_lastsp != self._const.LoopState.Open):
                self._update_log('FATAL:Opening FOFB loop...')
                self.run_callbacks('LoopState-Sel', self._const.LoopState.Open)
                self.run_callbacks('LoopState-Sts', self._const.LoopState.Open)
                self.set_loop_state(self._const.LoopState.Open, abort=True)

    # --- auxiliary corrector and fofbcontroller methods ---

    def _check_corr_connection(self):
        if self._corrs_dev.connected:
            return True
        self._update_log('ERR:Correctors not connected... aborted.')
        return False

    def _check_set_corrs_pwrstate(self):
        """Check and configure pwrstate.

        Control only correctors that are in the enable list.
        """
        self._update_log('Checking corrector pwrstate...')
        pwrstate = _Const.OffOn.On
        is_ok = 1 * (self._corrs_dev.pwrstate == pwrstate)
        idcs = _np.where((1 * self.corr_enbllist[:-1] - is_ok) > 0)[0]
        if idcs.size:
            self._update_log('Configuring corrector pwrstate...')
            self._corrs_dev.set_pwrstate(pwrstate, psindices=idcs)
            if self._corrs_dev.check_pwrstate(
                    pwrstate, psindices=idcs, timeout=5):
                self._update_log('...done!')
                return True
            self._update_log('ERR:Failed to set corrector pwrstate.')
            return False
        self._update_log('All ok.')
        return True

    def _check_set_corrs_opmode(self):
        """Check and configure opmode.

        Control only correctors that are in the enable list.
        """
        self._update_log('Checking corrector opmode...')
        opmode = self._corrs_dev.OPMODE_STS.manual \
            if self._loop_state == self._const.LoopState.Open \
            else self._corrs_dev.OPMODE_STS.fofb
        is_ok = 1 * (self._corrs_dev.opmode == opmode)
        idcs = _np.where((1 * self.corr_enbllist[:-1] - is_ok) > 0)[0]
        if idcs.size:
            self._update_log('Configuring corrector opmode...')
            self._corrs_dev.set_opmode(opmode, psindices=idcs)
            if self._corrs_dev.check_opmode(opmode, psindices=idcs, timeout=5):
                self._update_log('...done!')
                return True
            self._update_log('ERR:Failed to set corrector opmode.')
            return False
        self._update_log('All ok.')
        return True

    def _set_corrs_fofbacc_freeze(self):
        """Configure FOFBAccFreeze state.

        Keep in accordance with enable list and loop_state.
        """
        self._update_log('Setting corrector FOFCAccFreeze...')
        freeze = self._get_corrs_fofbacc_freeze_desired()
        self._corrs_dev.set_fofbacc_freeze(freeze)
        self._update_log('...done!')

    def _get_corrs_fofbacc_freeze_desired(self):
        if self._loop_state == self._const.LoopState.Open:
            freeze = _np.ones(len(self.corr_enbllist[:-1]))
        else:
            freeze = 1 * _np.logical_not(self.corr_enbllist[:-1])
        return freeze

    def _calc_corrs_coeffs(self, log=True):
        """Calculate corrector coefficients and gains."""
        if log:
            self._update_log('Calculating corrector coefficients ')
            self._update_log('and FOFB pre-accumulator gains...')

        # calculate coefficients and gains
        invmat = self._invrespmatconv[:-1]  # remove RF line
        coeffs = _np.zeros(invmat.shape)
        gains_mon = _np.zeros(self._const.nr_chcv)

        lgain_h, lgain_h_mon = self._loop_gain_h, self._loop_gain_mon_h
        lgain_v, lgain_v_mon = self._loop_gain_v, self._loop_gain_mon_v
        nrch, nrcv = self._const.nr_ch, self._const.nr_cv
        slch, slcv = slice(0, nrch), slice(nrch, nrch+nrcv)

        reso = self._const.ACCGAIN_RESO

        if self._invrespmat_normmode == self._const.GlobIndiv.Global:
            maxval = _np.amax(abs(invmat))

            gain_h_mon = _np.ceil(maxval * lgain_h_mon / reso) * reso
            gain_h = _np.ceil(maxval * lgain_h / reso) * reso
            if gain_h != 0:
                norm_h = gain_h / lgain_h
                coeffs[slch, :] = invmat[slch, :] / norm_h
                gains_mon[slch] = gain_h_mon * _np.ones(nrch)

            gain_v_mon = _np.ceil(maxval * lgain_v_mon / reso) * reso
            gain_v = _np.ceil(maxval * lgain_v / reso) * reso
            if gain_v != 0:
                norm_v = gain_v / lgain_v
                coeffs[slcv, :] = invmat[slcv, :] / norm_v
                gains_mon[slcv] = gain_v_mon * _np.ones(nrcv)

        elif self._invrespmat_normmode == self._const.GlobIndiv.Individual:
            maxval = _np.amax(abs(invmat), axis=1)

            gains_mon[slch] = _np.ceil(maxval[slch] * lgain_h_mon/reso) * reso
            gains_mon[slcv] = _np.ceil(maxval[slcv] * lgain_v_mon/reso) * reso

            gains = _np.zeros(self._const.nr_chcv)
            gains[slch] = _np.ceil(maxval[slch] * lgain_h / reso) * reso
            gains[slcv] = _np.ceil(maxval[slcv] * lgain_v / reso) * reso
            norm = _np.zeros(self._const.nr_chcv)
            norm[slch] = gains[slch] / lgain_h
            norm[slcv] = gains[slcv] / lgain_v
            idcs = norm > 0
            coeffs[idcs] = invmat[idcs] / norm[idcs][:, None]

        # handle FOFB BPM ordering
        nrbpm = self._const.nr_bpms
        coeffs[:, :nrbpm] = _np.roll(coeffs[:, :nrbpm], 1, axis=1)
        coeffs[:, nrbpm:] = _np.roll(coeffs[:, nrbpm:], 1, axis=1)

        # set internal states
        self._pscoeffs = coeffs
        self._psgains = gains_mon
        # update PVs
        self.run_callbacks('CorrCoeffs-Mon', list(self._pscoeffs.ravel()))
        self.run_callbacks('CorrGains-Mon', list(self._psgains.ravel()))

        if log:
            self._update_log('...done!')

    def _set_corrs_coeffs(self, log=True):
        """Set corrector coefficients and gains."""
        if log:
            self._update_log('Setting corrector gains and coefficients...')
        if not self._check_corr_connection():
            return False
        if not self._corrs_dev.check_invrespmat_row(self._pscoeffs):
            self._corrs_dev.set_invrespmat_row(self._pscoeffs)
        self._corrs_dev.set_fofbacc_gain(self._psgains)
        if log:
            self._update_log('...done!')
        return True

    def _check_fofbctrl_connection(self):
        if self._llfofb_dev.connected:
            return True
        self._update_log('ERR:FOFB Controllers not connected...')
        self._update_log('ERR:aborted.')
        return False

    def _update_fofbctrl_sync_enbllist(self):
        if self._fofbctrl_syncuseenbllist:
            bpmx = self._enable_lists['bpmx']
            bpmy = self._enable_lists['bpmy']
            mini = self._const.dccenbl_min
            dccenbl = _np.logical_or.reduce([bpmx, bpmy, mini])
            bpms = self._llfofb_dev.get_dccfmc_visible_bpms([
                self._const.bpm_names[i] for i, s in enumerate(dccenbl) if s])
            dccenbl = _np.array([b in bpms for b in self._const.bpm_names])
        else:
            dccenbl = _np.ones(self._const.nr_bpms, dtype=bool)
        self._fofbctrl_syncenbllist = dccenbl
        self.run_callbacks('CtrlrSyncEnblList-Mon', dccenbl)

    def _get_fofbctrl_bpmdcc_enbl(self):
        return [self._const.bpm_names[i] for i, s in
                enumerate(self._fofbctrl_syncenbllist) if s]

    def _do_fofbctrl_syncnet(self):
        bpms = self._get_fofbctrl_bpmdcc_enbl()
        self._update_log('Syncing FOFB net...')
        if self._llfofb_dev.cmd_sync_net(bpms=bpms):
            self._update_log('Sync sent to FOFB net.')
            return True
        self._update_log('ERR:Failed to sync FOFB net.')
        return False

    def _wait_fofbctrl_netsync(self):
        bpms = self._get_fofbctrl_bpmdcc_enbl()
        _t0 = _time.time()
        while _time.time() - _t0 < self._const.DEF_TIMEMINWAIT:
            if self._llfofb_dev.check_net_synced(bpms=bpms):
                self._update_log('Net synced, continuing...')
                return True
        self._update_log('ERR:Net not synced.')
        return False

    def _conf_fofbctrl_packetlossdetec(self):
        # disable enable status
        self._dsbl_fofbctrl_minbpmcnt_enbl()
        # set minimum packet count
        self._conf_fofbctrl_minbpmcnt()
        # return enable status
        self._conf_fofbctrl_minbpmcnt_enbl()

    def _dsbl_fofbctrl_minbpmcnt_enbl(self):
        timeout = self._const.DEF_TIMEWAIT
        self._update_log('Disabling packet loss detection...')
        self._llfofb_dev.set_min_bpm_count_enbl(0, timeout=timeout)
        self._update_log('...done!')
        return True

    def _conf_fofbctrl_minbpmcnt(self):
        timeout = self._const.DEF_TIMEWAIT
        count = int(_np.sum(self._fofbctrl_syncenbllist))
        if not _np.all(self._llfofb_dev.min_bpm_count == count):
            self._update_log('Setting MinBPMCnt PVs...')
            if self._llfofb_dev.set_min_bpm_count(count, timeout=timeout):
                self._update_log('...done!')
            else:
                self._update_log('ERR:Failed to sync MinBPMCnt.')
        else:
            self._update_log('MinBPMCnt PVs already configured.')
        return True

    def _conf_fofbctrl_minbpmcnt_enbl(self):
        timeout = self._const.DEF_TIMEWAIT
        sts = self._loop_packloss_detec_enbl
        if not _np.all(self._llfofb_dev.min_bpm_count_enbl == sts):
            self._update_log('Setting MinBPMCntEnbl PVs...')
            if self._llfofb_dev.set_min_bpm_count_enbl(sts, timeout=timeout):
                self._update_log('...done!')
            else:
                self._update_log('ERR:Failed to sync enable status.')
        else:
            self._update_log('MinBPMCntEnbl PVs already configured.')
        return True

    def _do_fofbctrl_reset(self):
        if self._thread_reset is not None and \
                self._thread_reset.is_alive():
            self._update_log('ERR: reset already in progress.')
            return False

        self._thread_reset = _epics.ca.CAThread(
            target=self._thread_fofbctrl_reset, daemon=True)
        self._thread_reset.start()
        return True

    def _thread_fofbctrl_reset(self):
        self._update_log('Reseting FOFB loop...')

        # disabling packet loss detection
        self._dsbl_fofbctrl_minbpmcnt_enbl()

        # reset interlock
        self._update_log('Sending reset to FOFB controllers...')
        if self._llfofb_dev.cmd_reset():
            self._update_log('...done!')
        else:
            self._update_log('ERR:Failed to reset controllers.')
            return

        # wait for packet count to be correct
        if not self._wait_fofbctrl_netsync():
            return

        # return packet loss detection to correct status
        self._update_log('Reconfiguring packet loss detection...')
        self._conf_fofbctrl_minbpmcnt_enbl()

    # --- auxiliary log methods ---

    def _update_log(self, msg):
        if 'ERR' in msg:
            _log.error(msg[4:])
        elif 'FATAL' in msg:
            _log.error(msg[6:])
        elif 'WARN' in msg:
            _log.warning(msg[5:])
        else:
            _log.info(msg)
        self.run_callbacks('Log-Mon', msg)

    def _check_corrs_configs(self):
        tplanned = 1.0/App.SCAN_FREQUENCY
        while not self.quit:
            if not self.scanning:
                _time.sleep(tplanned)
                continue

            _t0 = _time.time()

            # correctors status
            value = 0
            if self._corrs_dev.connected:
                idcs = _np.where(self.corr_enbllist[:-1] == 1)[0]
                # PwrStateOn
                state = self._const.OffOn.On
                if not self._corrs_dev.check_pwrstate(
                        state, psindices=idcs, timeout=0.2):
                    value = _updt_bit(value, 1, 1)
                # OpModeConfigured
                opmode = self._corrs_dev.OPMODE_STS.manual \
                    if self._loop_state == self._const.LoopState.Open \
                    else self._corrs_dev.OPMODE_STS.fofb
                if not self._corrs_dev.check_opmode(
                        opmode, psindices=idcs, timeout=0.2):
                    value = _updt_bit(value, 2, 1)
                # AccFreezeConfigured
                freeze = self._get_corrs_fofbacc_freeze_desired()
                if not self._corrs_dev.check_fofbacc_freeze(
                        freeze, timeout=0.2):
                    value = _updt_bit(value, 3, 1)
                # InvRespMatRowSynced
                if not self._corrs_dev.check_invrespmat_row(self._pscoeffs):
                    value = _updt_bit(value, 4, 1)
                # AccGainSynced
                if not self._corrs_dev.check_fofbacc_gain(self._psgains):
                    value = _updt_bit(value, 5, 1)
                # AccSatLimsSynced
                chn, chl = self._const.ch_names, self._ch_maxacccurr
                cvn, cvl = self._const.cv_names, self._cv_maxacccurr
                isok = self._corrs_dev.check_fofbacc_satmax(chl, psnames=chn)
                isok &= self._corrs_dev.check_fofbacc_satmin(-chl, psnames=chn)
                isok &= self._corrs_dev.check_fofbacc_satmax(cvl, psnames=cvn)
                isok &= self._corrs_dev.check_fofbacc_satmin(-cvl, psnames=cvn)
                if not isok:
                    value = _updt_bit(value, 6, 1)
                # AccDecimationSynced
                dec = self._corr_accdec_val
                if not self._corrs_dev.check_fofbacc_decimation(dec):
                    value = _updt_bit(value, 7, 1)
            else:
                value = 0b11111111

            self._corr_status = value
            self.run_callbacks('CorrStatus-Mon', self._corr_status)

            ttook = _time.time() - _t0
            tsleep = tplanned - ttook
            if tsleep > 0:
                _time.sleep(tsleep)
            else:
                _log.warning(
                    'Corrector configuration check took more than planned... '
                    '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))

    def _check_ctrls_configs(self):
        tplanned = 1.0/App.SCAN_FREQUENCY
        while not self.quit:
            if not self.scanning:
                _time.sleep(tplanned)
                continue

            _t0 = _time.time()

            # FOFB controllers status
            value = 0
            if self._llfofb_dev.connected:
                # BPMIdsConfigured
                if not self._llfofb_dev.bpm_id_configured:
                    value = _updt_bit(value, 1, 1)
                # NetSynced
                bpms = self._get_fofbctrl_bpmdcc_enbl()
                if not self._llfofb_dev.check_net_synced(bpms=bpms):
                    value = _updt_bit(value, 2, 1)
                # LinkPartnerConnected
                if not self._llfofb_dev.linkpartners_connected:
                    value = _updt_bit(value, 3, 1)
                # RefOrbSynced
                if not self._llfofb_dev.check_reforbx(self._reforbhw_x) or not\
                        self._llfofb_dev.check_reforby(self._reforbhw_y):
                    value = _updt_bit(value, 4, 1)
                # TimeFrameLenSynced
                tframelen = self._time_frame_len
                if not _np.all(self._llfofb_dev.time_frame_len == tframelen):
                    value = _updt_bit(value, 5, 1)
                # BPMLogTrigsConfigured
                if not self._llfofb_dev.bpm_trigs_configured:
                    value = _updt_bit(value, 6, 1)
                # OrbDistortionDetectionSynced
                sts = self._loop_max_orb_dist_enbl
                sts_ok = self._llfofb_dev.max_orb_distortion_enbl == sts
                thres = self._loop_max_orb_dist*self._const.CONV_UM_2_NM
                thres_ok = self._llfofb_dev.max_orb_distortion == thres
                if not _np.all(sts_ok) or not _np.all(thres_ok):
                    value = _updt_bit(value, 7, 1)
                # PacketLossDetectionSynced
                sts = self._loop_packloss_detec_enbl
                sts_ok = self._llfofb_dev.min_bpm_count_enbl == sts
                count = int(_np.sum(self._fofbctrl_syncenbllist))
                count_ok = self._llfofb_dev.min_bpm_count == count
                if not _np.all(sts_ok) or not _np.all(count_ok):
                    value = _updt_bit(value, 8, 1)
                # LoopInterlockOk
                if not self._llfofb_dev.interlock_ok:
                    value = _updt_bit(value, 9, 1)
                # SYSIDExcitationDisabled
                if not self._llfofb_dev.check_sysid_exc_disabled():
                    value = _updt_bit(value, 10, 1)
            else:
                value = 0b11111111111

            self._fofbctrl_status = value
            self.run_callbacks('CtrlrStatus-Mon', self._fofbctrl_status)

            ttook = _time.time() - _t0
            tsleep = tplanned - ttook
            if tsleep > 0:
                _time.sleep(tsleep)
            else:
                _log.warning(
                    'Controllers configuration check took more than planned... '
                    '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))



"""Epics PV derived classes."""

import time as _time
import numpy as _np

import epics as _epics

from ..envars import VACA_PREFIX as _prefix
from ..namesys import \
    SiriusPVName as _SiriusPVName, \
    get_pair_sprb as _get_pair_sprb


class EpicsProperty:
    """Pair of Epics PVs."""

    def __init__(self, name, prefix=_prefix, default_value=None,
                 connection_callback=None, callback=None):
        """Init."""
        try:
            self._name = _SiriusPVName(name)
            try:
                [self._pvname_sp, self._pvname_rb] = _get_pair_sprb(self._name)
            except TypeError:
                self._pvname_sp = name
                self._pvname_rb = name
        except Exception:
            self._name = name
            self._pvname_sp = name
            self._pvname_rb = name
        self._prefix = prefix
        self._default = default_value

        # Set callbacks
        self._connection_callback = connection_callback
        self._callback = callback
        callbacks = {'connection_callback': self._pv_connection_callback,
                     'callback': self._pv_callback}

        self._pv_sp = _epics.PV(
            self._prefix + ('-' if self._prefix else '') + self._pvname_sp,
            **callbacks)
        self._pv_rb = _epics.PV(
            self._prefix + ('-' if self._prefix else '') + self._pvname_rb,
            **callbacks)

    @property
    def name(self):
        """PV name."""
        return self._name

    @property
    def prefix(self):
        """PV prefix."""
        return self._prefix

    @property
    def suffix_sp(self):
        """PV SP suffix."""
        try:
            return self._pvname_sp.propty_suffix
        except Exception:
            return self._pvname_sp

    @property
    def suffix_rb(self):
        """PV RB suffix."""
        try:
            return self._pvname_rb.propty_suffix
        except Exception:
            return self._pvname_rb

    @property
    def pvname_sp(self):
        """Return SP pvname (without prefix)."""
        return self._pvname_sp

    @property
    def pvname_rb(self):
        """Return RB pvname (without prefix)."""
        return self._pvname_rb

    @property
    def connected(self):
        """State of connection."""
        return self._pv_sp.connected and self._pv_rb.connected

    @property
    def readback(self):
        """Property readback."""
        return self._pv_rb.get(timeout=0)

    @property
    def readback_string(self):
        """Property readback string."""
        return self._pv_rb.get(timeout=0, as_string=True)

    @property
    def setpoint(self):
        """Property setpoint."""
        return self._pv_sp.get(timeout=0)

    @property
    def setpoint_string(self):
        """Property setpoint string."""
        return self._pv_sp.get(timeout=0, as_string=True)

    @property
    def default(self):
        """Return default value."""
        return self._default

    @setpoint.setter
    def setpoint(self, value):
        """Set setpoint value."""
        if value is not None:
            self._pv_sp.value = value

    def set_setpoint_check(self, value, timeout):
        """Set setpoint value and check readback."""
        # setpoint
        self._pv_sp.value = value
        # check
        time0 = _time.time()
        while True:
            if self._pv_rb.value == value:
                return True
            if _time.time() - time0 > timeout:
                return False
            _time.sleep(min(0.1, timeout))

    def reset_default(self):
        """Reset to default value."""
        if self._default is not None:
            self.setpoint = self._default

    def set_callback(self, callback):
        """Set callback."""
        self._callback = callback

    def set_connection_callback(self, connection_callback):
        """Set connection callback."""
        self._connection_callback = connection_callback

    def _pv_connection_callback(self, **kwargs):
        if self._connection_callback is not None:
            kwargs['property'] = self
            self._connection_callback(**kwargs)

    def _pv_callback(self, **kwargs):
        if self._callback is not None:
            kwargs['property'] = self
            self._callback(**kwargs)


class EpicsPropertiesList:
    """List of Epics properties."""

    def __init__(self, properties):
        """Init."""
        self._properties = dict()
        for ppty in properties:
            self._properties[ppty.name] = ppty
        self._default = {p.name: p.default for p in self._properties.values()}
        self._logger = None

    @property
    def logger(self):
        """Return logger."""
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    @property
    def connected(self):
        """State of connection."""
        for ppty in self._properties.values():
            if not ppty.connected:
                return False
        return True

    @property
    def disconnected_properties(self):
        """Return list of disconnected properties."""
        props = []
        for name, ppty in self._properties.items():
            if not ppty.connected:
                props.append(name)
        return sorted(props)

    @property
    def properties(self):
        """Properties."""
        return sorted(self._properties.keys())

    @property
    def default(self):
        """Return default values dict."""
        return self._default

    @property
    def readbacks(self):
        """Return dict with all readbacks."""
        readbacks = dict()
        for name, ppty in self._properties.items():
            readbacks[name] = ppty.readback
        return readbacks

    def get_readback(self, name):
        """Return readback value of a property."""
        ppty = self._properties[name]
        return ppty.readback

    def get_readback_string(self, name):
        """Return readback value of a property as string."""
        ppty = self._properties[name]
        return ppty.readback_string

    def get_setpoint(self, name):
        """Return setpoint value of a property."""
        ppty = self._properties[name]
        return ppty.setpoint

    def get_setpoint_string(self, name):
        """Return setpoint value of a property as string."""
        ppty = self._properties[name]
        return ppty.setpoint_string

    def set_setpoints_check(self, setpoints, desired_readbacks=None,
                            timeout=5, order=None, rel_tol=1e-6, abs_tol=0.0):
        """Set setpoints of properties."""
        # setpoints
        if order is None:
            order = list(setpoints.keys())
        is_nok = list()
        for name in order:
            value = setpoints[name]
            if value is not None:
                ppty = self._properties[name]
                ppty.setpoint = value
                if 'Cmd' not in name:
                    is_nok.append(name)
            if self._logger is not None:
                self._logger.update(name)
        # check
        if desired_readbacks is None:
            desired_readbacks = setpoints
        time0 = _time.time()
        while _time.time() - time0 < timeout:
            finished = True
            for pvname, value in desired_readbacks.items():
                if value is None:
                    continue
                if pvname not in is_nok:
                    continue
                rbv = self._properties[pvname].readback
                if isinstance(value, (tuple, list, _np.ndarray)):
                    if not isinstance(rbv, (tuple, list, _np.ndarray)):
                        finished = False
                        break
                    if len(value) != len(rbv):
                        finished = False
                        break
                    if not all(_np.isclose(rbv, value,
                                           rtol=rel_tol, atol=abs_tol)):
                        finished = False
                        break
                else:
                    if not _np.isclose(rbv, value, rtol=rel_tol, atol=abs_tol):
                        finished = False
                        break
                if finished:
                    is_nok.remove(pvname)
            if finished:
                break
            _time.sleep(timeout/10.0)
        return finished, is_nok

    def reset_default(self):
        """Reset properties to default values."""
        for ppty in self._properties.values():
            ppty.reset_default()

    def set_callback(self, callback):
        """Set callback."""
        for ppty in self._properties.values():
            ppty.set_callback(callback)

    def set_connection_callback(self, connection_callback):
        """Set connection callback."""
        for ppty in self._properties.values():
            ppty.set_connection_callback(connection_callback)

    def __getitem__(self, key):
        """Property item."""
        if isinstance(key, int):
            properties = self.properties
            key = properties[key]
        return self._properties[key]



# #!/usr/bin/env python-sirius
# """BSMP class tests."""
# import unittest
# from unittest.mock import Mock
#
# from siriuspy.util import check_public_interface_namespace
#
#
# class _TestBSMP(TestCase):
#     """Test BSMP class."""
#
#     api = (
#         'ID_device',
#         'variables',
#         'functions',
#         'parse_stream',
#     )
#
#     def test_api(self):
#         """Test API."""
#         self.assertTrue(check_public_interface_namespace(BSMP, TestBSMP.api))
#
#     def test_init(self):
#         """Test parameters are initialized correctly."""
#         id_device = 0
#         variables = {'var1': 'val1', 'var2': 'val2'}
#         functions = {'func1': print, 'func2': BSMP.parse_stream}
#
#         bsmp = BSMP(id_device, variables, functions)
#
#         self.assertEqual(bsmp.ID_device, 0)
#         self.assertEqual(bsmp.variables, variables)
#         self.assertEqual(bsmp.functions, functions)
#
#     def test_parse_stream_small_stream(self):
#         """Test ValueError is raised when stream is too small."""
#         with self.assertRaises(ValueError):
#             BSMP.parse_stream(['\x30', '\x00', '\x25', '\x00'])
#
#     def test_parse_stream_no_checksum(self):
#         """Test ValueError is raised when checksum fails."""
#         stream = ['\x30', '\x00', '\x25', '\x00', '\x00', '\x00']
#         with self.assertRaises(ValueError):
#             BSMP.parse_stream(stream)
#
#     def test_parse_stream(self):
#         """Test stream is correctly parsed."""
#         stream = ['\x00', '\x0A', '\x02', '\x00', '\x00', '\x00']
#         stream = StreamChecksum.includeChecksum(stream)
#         id_receiver, id_cmd, load_size, load = BSMP.parse_stream(stream)
#         # print(id_receiver, id_cmd, load_size, load)
#         self.assertEqual(id_receiver, '\x00')
#         self.assertEqual(id_cmd, 10)
#         self.assertEqual(load_size, 2)
#         self.assertEqual(load, ['\x00', '\x00'])
#
#
# class _TestBSMPQuery(TestCase):
#     """Test BSMPQuery class."""
#
#     api = (
#         'slaves',
#         'add_slave',
#         'cmd_0x00',
#         'cmd_0x10',
#         'cmd_0x12',
#         'cmd_0x30',
#         'cmd_0x32',
#         'cmd_0x50'
#     )
#
#     def setUp(self):
#         """Common setup for all tests."""
#         self.variables = {'var1': 'val1', 'var2': 'val2'}
#         self.functions = {'func1': print, 'func2': BSMP.parse_stream}
#         self.slaves = list()
#         for i in range(3):
#             mock = Mock()
#             id_device = PropertyMock(return_value=i+1)
#             type(mock).ID_device = id_device
#             self.slaves.append(mock)
#
#         self.bsmp = BSMPQuery(self.variables, self.functions, self.slaves)
#
#     def test_api(self):
#         """Test API."""
#         self.assertTrue(
#             check_public_interface_namespace(BSMPQuery, TestBSMPQuery.api))
#
#     def test_init(self):
#         """Test initial values passed in the constructor."""
#         self.assertEqual(self.bsmp.ID_device, 0)
#         self.assertEqual(self.bsmp.variables, self.variables)
#         self.assertEqual(self.bsmp.functions, self.functions)
#
#     def test_cmd_0x00(self):
#         """Test query is called with correct parameters."""
#         self.bsmp.cmd_0x00(1)
#         self.bsmp.slaves[1].query.assert_called_with(0x00, ID_receiver=1)
#
#     def test_cmd_0x10(self):
#         """Test query is called with correct parameters."""
#         self.bsmp.cmd_0x10(1, 2)
#         self.bsmp.slaves[1].query.assert_called_with(
#             0x10, ID_receiver=1, ID_variable=2)
#
#     def test_cmd_0x12(self):
#         """Test query is called with correct parameters."""
#         self.bsmp.cmd_0x12(2, 1)
#         self.bsmp.slaves[2].query.assert_called_with(
#             0x12, ID_receiver=2, ID_group=1)
#
#     def test_cmd_0x30(self):
#         """Test query is called with correct parameters."""
#         self.bsmp.cmd_0x30(3, 4, 2)
#         self.bsmp.slaves[3].query.assert_called_with(
#             0x30, ID_receiver=3, ID_group=4, IDs_variable=2)
#
#     def test_cmd_0x30_exc(self):
#         """Test query is called with correct parameters."""
#         with self.assertRaises(ValueError):
#             self.bsmp.cmd_0x30(3, 2, 2)
#
#     def test_cmd_0x32(self):
#         """Test query is called with correct parameters."""
#         self.bsmp.cmd_0x32(3)
#         self.bsmp.slaves[3].query.assert_called_with(0x32, ID_receiver=3)
#
#     def test_cmd_0x50(self):
#         """Test query is called with correct parameters."""
#         self.bsmp.cmd_0x50(ID_receiver=1)
#         self.bsmp.slaves[1].query.assert_called_with(0x50, ID_receiver=1)
#
#
# class _TestBSMPResponse(TestCase):
#     """Test BSMPResponse class."""
#
#     api = (
#         'query',
#         'create_group_of_variables',
#         'remove_groups',
#         'cmd_0x01',
#         'cmd_0x11',
#         'cmd_0x13',
#         'cmd_0x51'
#     )
#
#     def test_api(self):
#         """Test API."""
#         self.assertTrue(
#             check_public_interface_namespace(
#                 BSMPResponse, TestBSMPResponse.api))
#
#     def test_query_2_resp(self):
#         """Test mapping from query to response is correct."""
#         self.assertEqual(BSMPResponse._query2resp[0x00], 'cmd_0x01')
#         self.assertEqual(BSMPResponse._query2resp[0x10], 'cmd_0x11')
#         self.assertEqual(BSMPResponse._query2resp[0x12], 'cmd_0x13')
#         self.assertEqual(BSMPResponse._query2resp[0x30], 'create_group_of_variables')
#         self.assertEqual(BSMPResponse._query2resp[0x32], 'remove_groups')
#         self.assertEqual(BSMPResponse._query2resp[0x50], 'cmd_0x51')



#!/usr/local/bin/python-sirius
"""PU Diag PVs."""

from ...namesys import SiriusPVName as _PVName
from ...pwrsupply.csdev import Const as _PSConst


class PUDiffPV:
    """Diff of a PU voltage setpoint and a monitor."""

    VOLTAGE_SP = 0
    VOLTAGE_MON = 1

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute difference between SP and Mon voltage values."""
        disconnected = \
            not computed_pv.pvs[PUDiffPV.VOLTAGE_SP].connected or \
            not computed_pv.pvs[PUDiffPV.VOLTAGE_MON].connected
        if disconnected:
            return None

        value_sp = computed_pv.pvs[PUDiffPV.VOLTAGE_SP].value
        value_rb = computed_pv.pvs[PUDiffPV.VOLTAGE_MON].value
        diff = value_rb - value_sp
        return {'value': diff}


class PUStatusPV:
    """Pulsed Power Supply Status PV."""

    BIT_PUCONNECT = 0b00001
    BIT_PWRSTATON = 0b00010
    BIT_PULSEISON = 0b00100
    BIT_VOLTGDIFF = 0b01000
    BIT_INTERLOCK = 0b10000

    PWRST_STS = 0
    PULSE_STS = 1
    DIFFVOLTG = 2
    INTRLCK_1 = 3
    INTRLCK_2 = 4
    INTRLCK_3 = 5
    INTRLCK_4 = 6
    INTRLCK_5 = 7
    INTRLCK_6 = 8
    INTRLCK_7 = 9
    INTRLCK_8 = 10

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute PU Status PV."""
        puname = _PVName(computed_pv.pvs[0].pvname).device_name
        value = 0
        # connected?
        disconnected = \
            not computed_pv.pvs[PUStatusPV.PWRST_STS].connected or \
            not computed_pv.pvs[PUStatusPV.PULSE_STS].connected or \
            not computed_pv.pvs[PUStatusPV.DIFFVOLTG].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_1].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_2].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_3].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_4].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_5].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_6].connected or \
            not computed_pv.pvs[PUStatusPV.INTRLCK_7].connected
        if 'Sept' not in puname:
            disconnected |= not computed_pv.pvs[PUStatusPV.INTRLCK_8].connected
        if disconnected:
            value |= PUStatusPV.BIT_PUCONNECT
            value |= PUStatusPV.BIT_PWRSTATON
            value |= PUStatusPV.BIT_PULSEISON
            value |= PUStatusPV.BIT_VOLTGDIFF
            value |= PUStatusPV.BIT_INTERLOCK
            return {'value': value}

        # pwrstate?
        pwrsts = computed_pv.pvs[PUStatusPV.PWRST_STS].value
        if pwrsts != _PSConst.OffOn.On or pwrsts is None:
            value |= PUStatusPV.BIT_PWRSTATON
        else:
            # voltage-diff?
            severity = computed_pv.pvs[PUStatusPV.DIFFVOLTG].severity
            if severity != 0:
                value |= PUStatusPV.BIT_VOLTGDIFF

        # pulse?
        pulsests = computed_pv.pvs[PUStatusPV.PULSE_STS].value
        if pulsests != _PSConst.OffOn.On or pulsests is None:
            value |= PUStatusPV.BIT_PULSEISON

        # interlocks?
        intlk1 = computed_pv.pvs[PUStatusPV.INTRLCK_1].value
        intlk2 = computed_pv.pvs[PUStatusPV.INTRLCK_2].value
        intlk3 = computed_pv.pvs[PUStatusPV.INTRLCK_3].value
        intlk4 = computed_pv.pvs[PUStatusPV.INTRLCK_4].value
        intlk5 = computed_pv.pvs[PUStatusPV.INTRLCK_5].value
        intlk6 = computed_pv.pvs[PUStatusPV.INTRLCK_6].value
        intlk7 = computed_pv.pvs[PUStatusPV.INTRLCK_7].value
        cond = intlk1 != 1 or intlk1 is None
        cond |= intlk2 != 1 or intlk2 is None
        cond |= intlk3 != 1 or intlk3 is None
        cond |= intlk4 != 1 or intlk4 is None
        cond |= intlk5 != 1 or intlk5 is None
        cond |= intlk6 != 1 or intlk6 is None
        cond |= intlk7 != 1 or intlk7 is None
        if 'Sept' not in puname:
            intlk8 = computed_pv.pvs[PUStatusPV.INTRLCK_8].value
            cond |= intlk8 != 1 or intlk8 is None
        if cond:
            value |= PUStatusPV.BIT_INTERLOCK

        return {'value': value}



"""Epics-related subpackage."""

# the following parameter is used to establish connections with IOC PVs.
CONNECTION_TIMEOUT = 0.050  # [s]
GET_TIMEOUT = 5.0  # [s]

from .pv import PV
from .pv_time_serie import *
from .properties import *
from .multiproc import CAProcessSpawn
from .threading import CAThread

del pv
del pv_time_serie
del properties
del multiproc
del threading
